#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import numbers

import numpy

import awkward.array.base

class IndexedArray(awkward.array.base.AwkwardArray):
    def __init__(self, index, content):
        raise NotImplementedError

# class IndexedArray(awkward.array.base.AwkwardArray):
#     def __init__(self, index, content):
#         self.index = index
#         self.content = content

#     @property
#     def index(self):
#         return self._index

#     @index.setter
#     def index(self, value):
#         value = self._toarray(value, self.INDEXTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray))

#         if len(value.shape) != 1:
#             raise TypeError("index must have 1-dimensional shape")
#         if value.shape[0] == 0:
#             value = value.view(self.INDEXTYPE)
#         if not issubclass(value.dtype.type, numpy.integer):
#             raise TypeError("index must have integer dtype")

#         self._index = value

#     @property
#     def content(self):
#         return self._content

#     @content.setter
#     def content(self, value):
#         self._content = self._toarray(value, self.CHARTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray))

#     @property
#     def dtype(self):
#         return self._content.dtype

#     @property
#     def shape(self):
#         return (len(self._index),) + self._content.shape[1:]

#     def __len__(self):
#         return len(self._index)
        
#     def __getitem__(self, where):
#         if self._isstring(where):
#             return IndexedArray(self._index, self._content[where])

#         return self._content[self._index[where]]

class ByteIndexedArray(IndexedArray):
    def __init__(self, index, content, dtype):
        raise NotImplementedError
        
# class ByteIndexedArray(IndexedArray):
#     def __init__(self, index, content, dtype):
#         super(ByteIndexedArray, self).__init__(index, content)
#         self.dtype = dtype

#     @property
#     def content(self):
#         return self._content

#     @content.setter
#     def content(self, value):
#         self._content = self._toarray(value, self.CHARTYPE, numpy.ndarray).view(self.CHARTYPE).reshape(-1)

#     @property
#     def dtype(self):
#         return self._dtype

#     @dtype.setter
#     def dtype(self, value):
#         self._dtype = numpy.dtype(value)
        
#     def __getitem__(self, where):
#         if self._isstring(where):
#             return ByteIndexedArray(self._index, self._content[where], self._dtype)

#         starts = self._index[where]

#         if len(starts.shape) == 0:
#             pos, offset = divmod(starts, self._dtype.itemsize)
#             return numpy.frombuffer(self._content, dtype=self._dtype, count=(pos + 1), offset=offset)[pos]

#         else:
#             if len(starts) == 0:
#                 return numpy.empty(0, dtype=self._dtype)

#             else:
#                 hold = numpy.empty(len(starts), dtype=self._dtype)

#                 contidx = numpy.empty(len(starts) * self._dtype.itemsize, dtype=self.INDEXTYPE)
#                 contidx[::self._dtype.itemsize] = starts
#                 for offset in range(1, self._dtype.itemsize):
#                     contidx[offset::self._dtype.itemsize] = contidx[::self._dtype.itemsize] + offset
                
#                 holdidx = numpy.empty(len(starts) * self._dtype.itemsize, dtype=self.INDEXTYPE)
#                 holdidx[::self._dtype.itemsize] = numpy.arange(0, len(starts) * self._dtype.itemsize, self._dtype.itemsize)
#                 for offset in range(1, self._dtype.itemsize):
#                     holdidx[offset::self._dtype.itemsize] = holdidx[::self._dtype.itemsize] + offset

#                 numpy.frombuffer(hold, dtype=self.CHARTYPE)[holdidx] = self._content[contidx]
#                 return hold

class IndexedMaskedArray(IndexedArray):
    def __init__(self, index, content, maskedwhen=-1):
        raise NotImplementedError

# class IndexedMaskedArray(IndexedArray):
#     def __init__(self, index, content, maskedwhen=-1):
#         super(IndexedMaskedArray, self).__init__(index, content)
#         self.maskedwhen = maskedwhen

#     @property
#     def maskedwhen(self):
#         return self._maskedwhen

#     @maskedwhen.setter
#     def maskedwhen(self, value):
#         if not isinstance(value, (numbers.Integral, numpy.integer)):
#             raise TypeError("maskedwhen must be an integer")
#         self._maskedwhen = value

#     def __getitem__(self, where):
#         if self._isstring(where):
#             return IndexedMaskedArray(self._index, self._content[where], maskedwhen=self._maskedwhen)

#         if not isinstance(where, tuple):
#             where = (where,)
#         head, tail = where[0], where[1:]

#         if isinstance(head, (numbers.Integral, numpy.integer)):
#             if self._index[head] == self._maskedwhen:
#                 return numpy.ma.masked
#             else:
#                 return self._content[self._singleton((self._index[head],) + tail)]
#         else:
#             return IndexedMaskedArray(self._index[head], self._content[self._singleton((slice(None),) + tail)], maskedwhen=self._maskedwhen)

class UnionArray(awkward.array.base.AwkwardArray):
    def __init__(self, tags, index, contents):
        raise NotImplementedError

# class UnionArray(awkward.array.base.AwkwardArray):
#     @classmethod
#     def fromtags(cls, tags, contents):
#         raise NotImplementedError

#     def __init__(self, tags, index, contents):
#         self.tags = tags
#         self.index = index
#         self.contents = contents

#     @property
#     def tags(self):
#         return self._tags

#     @tags.setter
#     def tags(self, value):
#         value = self._toarray(value, self.INDEXTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray))

#         if len(value.shape) != 1:
#             raise TypeError("tags must have 1-dimensional shape")
#         if value.shape[0] == 0:
#             value = value.view(self.INDEXTYPE)
#         if not issubclass(value.dtype.type, numpy.integer):
#             raise TypeError("tags must have integer dtype")

#         self._tags = value

#     @property
#     def index(self):
#         return self._index

#     @index.setter
#     def index(self, value):
#         value = self._toarray(value, self.INDEXTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray))

#         if len(value.shape) != 1:
#             raise TypeError("index must have 1-dimensional shape")
#         if value.shape[0] == 0:
#             value = value.view(self.INDEXTYPE)
#         if not issubclass(value.dtype.type, numpy.integer):
#             raise TypeError("index must have integer dtype")

#         self._index = value

#     @property
#     def contents(self):
#         return self._contents

#     @contents.setter
#     def contents(self, value):
#         self._contents = tuple(self._toarray(x, self.CHARTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray)) for x in value)

#     @property
#     def dtype(self):
#         return numpy.dtype(object)

#     @property
#     def shape(self):
#         return (len(self._tags),)

#     def __len__(self):
#         return len(self._tags)

#     def __getitem__(self, where):
#         if self._isstring(where):
#             return UnionArray(self._tags, self._index, tuple(x[where] for x in self._contents))

#         if self._tags.shape != self._index.shape:
#             raise ValueError("tags shape ({0}) does not match index shape ({1})".format(self._tags.shape, self._index.shape))

#         if not isinstance(where, tuple):
#             where = (where,)
#         head, tail = where[0], where[1:]

#         tags = self._tags[head]
#         index = self._index[head]
#         assert tags.shape == index.shape

#         uniques = numpy.unique(tags)
#         if len(uniques) == 1:
#             return self._contents[uniques[0]][self._singleton((index,) + tail)]
#         else:
#             return UnionArray(tags, index, self._contents)



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import functools
import numbers
import types

import awkward.array.base
import awkward.type
import awkward.util

class Table(awkward.array.base.AwkwardArray):
    ##################### class Row

    class Row(object):
        __slots__ = ["_table", "_index"]

        def __init__(self, table, index):
            self._table = table
            self._index = index

        def __repr__(self):
            return "<{0} {1}>".format(self._table.rowname, self._index)

        def __hasattr__(self, name):
            return name in self._table._content or "_" + name in self._table._content

        def __contains__(self, name):
            return name in self._table._content

        def __getattr__(self, name):
            if name == "tolist":
                return lambda: dict((n, self._table._try_tolist(x[self._index])) for n, x in self._table._content.items())

            content = self._table._content.get(name, None)
            if content is not None:
                return content[self._index]

            content = self._table._content.get("_" + name, None)
            if content is not None:
                return content[self._index]

            raise AttributeError("neither {0} nor _{1} are columns in this {2}".format(name, name, self._table.rowname))

        def __getitem__(self, where):
            if isinstance(where, awkward.util.string):
                return self._table._content[where][self._index]

            elif awkward.util.isstringslice(where):
                table = self._table.copy(content=awkward.util.OrderedDict([(n, self._table._content[n]) for n in where]))
                return table.Row(table, self._index)

            else:
                index = self._index
                if not isinstance(index, tuple):
                    index = (index,)
                if not isinstance(where, tuple):
                    where = (where,)
                return self._table.Row(table, index + where)

        def __dir__(self):
            return ["_" + x if len(x) == 0 or not x[0].isalpha() else x for x in self._table._content]

        def __iter__(self):
            i = 0
            while str(i) in self._table._content:
                yield self._table._content[str(i)]
                i += 1

        def __len__(self):
            i = 0
            while str(i) in self._table._content:
                i += 1
            return i

        def __eq__(self, other):
            if not isinstance(other, Table.Row):
                return False
            elif self._table is other._table and self._index == other._index:
                return True
            else:
                return set(self._table._content) == set(other._table._content) and all(self._table._content[n][self._index] == other._table._content[n][other._index] for n in self._table._content)

        def __ne__(self, other):
            return not self.__eq__(other)

    ##################### class Table

    def __init__(self, columns1={}, *columns2, **columns3):
        self._view = None
        self._base = None
        self.rowname = "Row"
        self._content = awkward.util.OrderedDict()

        seen = set()
        if isinstance(columns1, dict):
            for n, x in columns1.items():
                if n in seen:
                    raise ValueError("column {0} occurs more than once".format(repr(n)))
                seen.add(n)

                self[n] = x

            if len(columns2) != 0:
                raise TypeError("only one positional argument when first argument is a dict")

        else:
            self["0"] = columns1
            for i, x in enumerate(columns2):
                self[str(i + 1)] = x

        seen.update(self._content)

        for n, x in columns3.items():
            if n in seen:
                raise ValueError("column {0} occurs more than once".format(repr(n)))
            seen.add(n)

            self[n] = x

    @classmethod
    def named(cls, rowname, columns1={}, *columns2, **columns3):
        out = cls(columns1, *columns2, **columns3)
        out.rowname = rowname
        return out

    @property
    def rowname(self):
        return self._rowname

    @rowname.setter
    def rowname(self, value):
        if not isinstance(value, awkward.util.string):
            raise TypeError("rowname must be a string")
        self._rowname = value

    @classmethod
    def fromrec(cls, recarray):
        if not isinstance(recarray, awkward.util.numpy.ndarray) or recarray.dtype.names is None:
            raise TypeError("recarray must be a Numpy structured array")
        out = cls()
        for n in recarray.dtype.names:
            out[n] = recarray[n]
        return out

    def copy(self, content=None):
        out = self.__class__.__new__(self.__class__)
        out._view = self._view
        out._base = self._base
        out._rowname = self._rowname
        out._content = self._content
        if content is not None and isinstance(content, dict):
            out._content = awkward.util.OrderedDict(content.items())
        elif content is not None:
            out._content = awkward.util.OrderedDict(content)
        else:
            out._content = awkward.util.OrderedDict(self._content.items())
        return out

    def deepcopy(self, content=None):
        out = self.copy(content=content)
        index = out._index()
        if index is None:
            out._content = awkward.util.OrderedDict([(n, awkward.util.deepcopy(x)) for n, x in out._content.items()])
        else:
            out._content = awkward.util.OrderedDict([(n, awkward.util.deepcopy(x[index])) for n, x in out._content.items()])
            out._view = None
            out._base = None
        return out

    def empty_like(self, **overrides):
        out = self.__class__.__new__(self.__class__)
        out._view = None
        out._base = None
        out._rowname = self._rowname
        out._content = awkward.util.OrderedDict()
        return out

    def zeros_like(self, **overrides):
        out = self.empty_like(**overrides)
        for n, x in self._content.items():
            if isinstance(x, awkward.util.numpy.ndarray):
                out[n] = awkward.util.numpy.zeros_like(x)
            else:
                out[n] = x.zeros_like(**overrides)
        return out

    def ones_like(self, **overrides):
        out = self.empty_like(**overrides)
        for n, x in self._content.items():
            if isinstance(x, awkward.util.numpy.ndarray):
                out[n] = awkward.util.numpy.ones_like(x)
            else:
                out[n] = x.ones_like(**overrides)
        return out

    @property
    def columns(self):
        return [x for x in self._content if awkward.util.isintstring(x)] + [x for x in self._content if awkward.util.isidentifier(x)]

    @property
    def allcolumns(self):
        return list(self._content)

    @property
    def base(self):
        return self._base

    def _argfields(self, function):
        if not isinstance(function, types.FunctionType):
            raise TypeError("function (or lambda) required")

        required = function.__code__.co_varnames[:function.__code__.co_argcount]
        has_varargs = (function.__code__.co_flags & 0x04) != 0
        has_kwargs = (function.__code__.co_flags & 0x08) != 0

        args = []
        kwargs = {}

        order = self.columns

        for i, n in enumerate(required):
            if n in self._content:
                args.append(n)
            elif str(i) in self._content:
                args.append(str(i))
            else:
                args.append(order[i])

        if has_varargs:
            while str(i) in self._content:
                args.append(str(i))
                i += 1

        if has_kwargs:
            kwargs = [n for n in self._content if n not in required]

        return args, kwargs

    @property
    def dtype(self):
        return awkward.util.numpy.dtype([(n, x.dtype) for n, x in self._content.items()])

    def _length(self):
        if self._view is None:
            if len(self._content) == 0:
                return 0
            else:
                return min([len(x) for x in self._content.values()])

        elif isinstance(self._view, tuple):
            start, step, length = self._view
            return length

        else:
            return len(self._view)

    def _index(self):
        if self._view is None:
            return None

        elif isinstance(self._view, tuple):
            start, step, length = self._view
            stop = start + step*length
            if stop < 0:
                stop = None
            return slice(start, stop, step)

        else:
            return self._view

    def _newslice(self, head):
        if isinstance(head, (numbers.Integral, awkward.util.numpy.integer)):
            original_head = head

            if self._view is None:
                length = self._length()
                if head < 0:
                    head += length
                if not 0 <= head < length:
                    IndexError("index {0} out of bounds for length {1}".format(original_head, length))
                return head

            elif isinstance(self._view, tuple):
                mystart, mystep, mylength = self._view
                if head < 0:
                    head += mylength
                if not 0 <= head < mylength:
                    IndexError("index {0} out of bounds for length {1}".format(original_head, mylength))
                return mystart + mystep*head

            else:
                length = len(self._view)
                if head < 0:
                    head += length
                if not 0 <= head < length:
                    IndexError("index {0} out of bounds for length {1}".format(original_head, length))
                return self._view[head]

        elif isinstance(head, slice):
            if self._view is None or isinstance(self._view, tuple):
                start, stop, step = head.indices(self._length())
                if step == 0:
                    raise ValueError("slice step cannot be zero")
                if (step > 0 and stop - start > 0) or (step < 0 and stop - start < 0):
                    d, m = divmod(abs(start - stop), abs(step))
                    length = d + (1 if m != 0 else 0)
                else:
                    length = 0

            if self._view is None:
                return start, step, length

            elif isinstance(self._view, tuple):
                mystart, mystep, mylength = self._view
                if step > 0:
                    skip = start
                else:
                    skip = mylength - 1 - start
                return mystart + mystep*start, mystep*step, min(mylength - skip, length)

            else:
                return self._view[head]

        else:
            head = awkward.util.toarray(head, awkward.util.INDEXTYPE, (awkward.util.numpy.ndarray, awkward.array.base.AwkwardArray))
            if issubclass(head.dtype.type, awkward.util.numpy.integer):
                length = self._length()
                negative = (head < 0)
                head[negative] += length
                if not awkward.util.numpy.bitwise_and(0 <= head, head < length).all():
                    raise IndexError("some indexes out of bounds for length {0}".format(length))

                if self._view is None:
                    return head

                elif isinstance(self._view, tuple):
                    mystart, mystep, mylength = self._view
                    return awkward.util.numpy.arange(mystart, mystart + mystep*mylength, mystep)[head]

                else:
                    return self._view[head]

            elif issubclass(head.dtype.type, (awkward.util.numpy.bool, awkward.util.numpy.bool_)):
                length = self._length()
                if len(head) != length:
                    raise IndexError("boolean index of length {0} does not fit array of length {1}".format(len(head), length))

                if self._view is None:
                    return awkward.util.numpy.arange(length)[head]

                elif isinstance(self._view, tuple):
                    mystart, mystep, mylength = self._view
                    return awkward.util.numpy.arange(mystart, mystart + mystep*mylength, mystep)[head]

                else:
                    return self._view[head]

            else:
                raise TypeError("cannot interpret dtype {0} as a fancy index or mask".format(head.dtype))

    @property
    def shape(self):
        return (self._length(),)
        
    @property
    def type(self):
        return awkward.type.ArrayType(self._length(), functools.reduce(lambda a, b: a & b, [awkward.type.ArrayType(n, awkward.type.fromarray(x).to) for n, x in self._content.items()]))

    def __len__(self):
        return self._length()

    def _valid(self):
        return True

    def __iter__(self):
        if self._view is None:
            length = self._length()
            i = 0
            while i < length:
                yield self.Row(self, i)
                i += 1

        elif isinstance(self._view, tuple):
            mystart, mystep, mylength = self._view
            i = 0
            while i < mylength:
                yield self.Row(self, mystart + i*mystep)
                i += 1
        else:
            for i in self._view:
                yield self.Row(self, i)

    def __getitem__(self, where):
        if awkward.util.isstringslice(where):
            if isinstance(where, awkward.util.string):
                index = self._index()
                if index is None:
                    return self._content[where]
                else:
                    return self._content[where][index]
            else:
                return self.copy(content=[(n, self._content[n]) for n in where])

        if isinstance(where, tuple) and where == ():
            return self
        if not isinstance(where, tuple):
            where = (where,)
        head, tail = where[0], where[1:]

        newslice = self._newslice(head)

        if isinstance(newslice, (numbers.Integral, awkward.util.numpy.integer)):
            return self.Row(self, newslice)

        else:
            out = self.copy(content=self._content)
            out._view = newslice
            out._base = self._base
            return out

    def __setitem__(self, where, what):
        if self._view is not None:
                raise ValueError("new columns can only be attached to the original table, not a view (try table.base['col'] = array)")

        if isinstance(where, awkward.util.string):
            self._content[where] = awkward.util.toarray(what, awkward.util.CHARTYPE, (awkward.util.numpy.ndarray, awkward.array.base.AwkwardArray))

        elif awkward.util.isstringslice(where):
            if len(where) != len(what):
                raise ValueError("number of keys ({0}) does not match number of provided arrays ({1})".format(len(where), len(what)))
            for x, y in zip(where, what):
                self._content[x] = awkward.util.toarray(y, awkward.util.CHARTYPE, (awkward.util.numpy.ndarray, awkward.array.base.AwkwardArray))

        else:
            raise TypeError("invalid index for assigning column to Table: {0}".format(where))

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method != "__call__":
            return NotImplemented

        inputsdict = None
        for x in inputs:
            if isinstance(x, Table):
                if inputsdict is None:
                    inputsdict = awkward.util.OrderedDict([(n, []) for n in x._content])
                    table = x
                elif set(inputsdict) != set(x._content):
                    raise ValueError("Tables have different sets of columns")

        assert inputsdict is not None

        for x in inputs:
            if isinstance(x, Table):
                index = x._index()
                for n, y in x._content.items():
                    if index is None:
                        inputsdict[n].append(y)
                    else:
                        inputsdict[n].append(y[index])
            else:
                for n in inputsdict:
                    inputsdict[n].append(x)

        newcolumns = {}
        tuplelen = None
        for n, x in inputsdict.items():
            newcolumns[n] = getattr(ufunc, method)(*x, **kwargs)

            if tuplelen is None:
                if isinstance(x, tuple):
                    tuplelen = len(x)
                else:
                    tuplelen = False
            elif isinstance(x, tuple) != tuplelen:
                raise AssertionError("ufuncs return tuples of different lengths or some tuples and some non-tuples")

        assert len(newcolumns) != 0
        assert tuplelen is not None

        if awkward.util.iscomparison(ufunc):
            out = None
            for x in newcolumns.values():
                assert isinstance(x, awkward.util.numpy.ndarray)
                assert issubclass(x.dtype.type, (awkward.util.numpy.bool_, awkward.util.numpy.bool))
                if out is None:
                    out = x
                else:
                    out = awkward.util.numpy.bitwise_and(out, x, out=out)
            assert out is not None
            return out

        if method == "at":
            return None

        if tuplelen is False:
            out = table.empty_like()
            for n in inputsdict:
                out[n] = newcolumns[n]
            return out

        else:
            out = [table.empty_like() for i in range(tuplelen)]
            for n in inputsdict:
                for i in range(tuplelen):
                    out[i][n] = newcolumns[n]
            return tuple(out)

    def tolist(self):
        return list(x.tolist() for x in self)
    
    def pandas(self):
        import pandas
        return pandas.DataFrame(self._content)



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numbers

import numpy

import awkward.util

def fromarray(array):
    if isinstance(array, numpy.ndarray):
        if array.dtype.names is None:
            out = ArrayType(*(array.shape + (array.dtype,)))
            if isinstance(array, numpy.ma.MaskedArray):
                out = OptionType(out)
            return out
        else:
            table = TableType.__new__(TableType)
            table._fields = awkward.util.OrderedDict()
            for n in array.dtype.names:
                table[n] = array.dtype[n]
            out = ArrayType(*(array.shape + (table,)))
            if isinstance(array, numpy.ma.MaskedArray):
                out = OptionType(out)
            return out
    else:
        return array.type

class Type(object):
    def __or__(self, other):
        out = UnionType.__new__(UnionType)

        if isinstance(self, UnionType) and isinstance(other, UnionType):
            out._possibilities = self._possibilities + other._possibilities
        elif isinstance(self, UnionType):
            out._possibilities = self._possibilities + [other]
        elif isinstance(other, UnionType):
            out._possibilities = [self] + other._possibilities
        else:
            out._possibilities = [self, other]

        return out

    def _labeled(self):
        seen = set()
        labeled = []
        def find(x):
            if isinstance(x, Type):
                if id(x) not in seen:
                    seen.add(id(x))
                    if isinstance(x, ArrayType):
                        find(x._to)
                    elif isinstance(x, TableType):
                        for y in x._fields.values():
                            find(y)
                    elif isinstance(x, UnionType):
                        for y in x._possibilities:
                            find(y)
                    elif isinstance(x, OptionType):
                        find(x._type)
                else:
                    labeled.append(x)
        find(self)
        return labeled

    def __repr__(self):
        return self._repr(self._labeled(), set())

    def _repr(self, labeled, seen):
        if id(self) in seen:
            for i, x in enumerate(labeled):
                if self is x:
                    return "T{0}".format(i)
        else:
            for i, x in enumerate(labeled):
                if self is x:
                    out = "T{0} := ".format(i)
                    break
            else:
                out = ""
            seen.add(id(self))
            return out + self._subrepr(labeled, seen)

    def __str__(self):
        return self._str(self._labeled(), set(), "")

    def _str(self, labeled, seen, indent):
        if id(self) in seen:
            for i, x in enumerate(labeled):
                if self is x:
                    return indent + "T{0}".format(i)
            raise AssertionError("{0} not in {1}".format(id(self), seen))
        else:
            for i, x in enumerate(labeled):
                if self is x:
                    out = indent + "T{0} := ".format(i)
                    break
            else:
                out = ""
            seen.add(id(self))
            return out + self._substr(labeled, seen, indent + (" " * len(out)))
        
    @staticmethod
    def _copy(x, seen):
        if id(x) in seen:
            return seen[id(x)]

        elif isinstance(x, ArrayType):
            seen[id(x)] = ArrayType.__new__(ArrayType)
            seen[id(x)]._takes = x.takes
            seen[id(x)]._to = Type._copy(x.to, seen)
            return seen[id(x)]

        elif isinstance(x, TableType):
            seen[id(x)] = TableType.__new__(TableType)
            seen[id(x)]._fields = awkward.util.OrderedDict()
            for n, y in x._fields.items():
                seen[id(x)]._fields[n] = Type._copy(y, seen)
            return seen[id(x)]

        elif isinstance(x, UnionType):
            seen[id(x)] = UnionType.__new__(UnionType)
            seen[id(x)]._possibilities = []
            for y in x._possibilities:
                seen[id(x)]._possibilities.append(Type._copy(y, seen))
            return seen[id(x)]

        elif isinstance(x, OptionType):
            seen[id(x)] = OptionType.__new__(OptionType)
            seen[id(x)]._type = Type._copy(x._type, seen)
            return seen[id(x)]

        else:
            return x

    @staticmethod
    def _canonical(x, seen):
        if id(x) not in seen:
            # apply string-integer commutation so that TableTypes are nested as deeply as possible
            if isinstance(x, TableType) and len(x._fields) > 0 and all(isinstance(y, ArrayType) for y in x._fields.values()):
                newtable = TableType.__new__(TableType)
                newtable._fields = awkward.util.OrderedDict()
                first = None
                for n, y in x._fields.items():
                    if first is None:
                        first = y._takes
                    elif first != y._takes:
                        break
                    newtable._fields[n] = y._to
                else:
                    x = ArrayType.__new__(ArrayType)
                    x._takes = first
                    x._to = newtable

            # apply union(X, union(Y)) == union(X, Y)
            if isinstance(x, UnionType) and any(isinstance(y, UnionType) for y in x._possibilities):
                possibilities = []
                for y in x._possibilities:
                    if isinstance(y, UnionType):
                        possibilities.extend(y._possibilities)
                    else:
                        possibilities.append(y)
                x = UnionType.__new__(UnionType)
                x._possibilities = possibilities

            # apply optional(optional(X)) == optional(X)
            if isinstance(x, OptionType) and isinstance(x._type, OptionType):
                x = x._type

            seen.add(id(x))
            if isinstance(x, ArrayType):
                x._to = Type._canonical(x._to, seen)
            elif isinstance(x, TableType):
                for n in x._fields:
                    x._fields[n] = Type._canonical(x._fields[n], seen)
            elif isinstance(x, UnionType):
                for i in range(len(x._possibilities)):
                    x._possibilities[i] = Type._canonical(x._possibilities[i], seen)
            elif isinstance(x, OptionType):
                x._type = Type._canonical(x._type, seen)

        return x

    def __eq__(self, other):
        one = Type._canonical(Type._copy(self, {}), set())
        two = Type._canonical(Type._copy(other, {}), set())
        return one._eq(two, set())

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def _finaltype(x):
        if isinstance(x, type) and issubclass(x, (numbers.Number, numpy.generic)):
            return numpy.dtype(x)
        elif isinstance(x, (awkward.util.unicode, bytes)):
            return numpy.dtype(x)
        else:
            return x

class ArrayType(Type):
    def __init__(self, *args):
        if len(args) == 0:
            raise ValueError("type specification missing")

        elif len(args) == 1:
            self.takes = None
            self.to = args[0]

        elif isinstance(args[0], awkward.util.string):
            self.__class__ = TableType
            self._fields = awkward.util.OrderedDict()
            if len(args) == 2:
                self[args[0]] = args[1]
            else:
                self[args[0]] = ArrayType(*args[1:])

        else:
            self.takes = args[0]
            if len(args) == 2:
                self.to = args[1]
            else:
                self.to = ArrayType(*args[1:])

    @property
    def takes(self):
        return self._takes

    @takes.setter
    def takes(self, value):
        if value == numpy.inf or (isinstance(value, (numbers.Integral, numpy.integer)) and value >= 0):
            self._takes = value

        else:
            raise ValueError("{0} is not allowed in type specification".format(value))

    @property
    def to(self):
        return self._to

    @to.setter
    def to(self, value):
        if isinstance(value, Type):
            self._to = value
        else:
            self._to = self._finaltype(value)

    def _subrepr(self, labeled, seen):
        if isinstance(self._to, Type):
            to = self._to._repr(labeled, seen)
            if to.startswith("ArrayType(") and to.endswith(")"):
                to = to[10:-1]
            return "ArrayType({0}, {1})".format(repr(self._takes), to)
        else:
            return "ArrayType({0}, {1})".format(repr(self._takes), repr(self._to))

    def _substr(self, labeled, seen, indent):
        takes = "[0, {0}) -> ".format(self._takes)
        if isinstance(self._to, Type):
            to = self._to._str(labeled, seen, indent + (" " * len(takes))).lstrip(" ")
        else:
            to = str(self._to)
        return takes + to

    def _eq(self, other, seen):
        if self is other:
            return True
        elif id(self) in seen:
            return False
        else:
            seen.add(id(self))
            return isinstance(other, ArrayType) and self._takes == other._takes and self._to == other._to

    def __hash__(self):
        return hash((ArrayType, self._takes, self._to))

class TableType(Type):
    def __init__(self, **fields):
        self._fields = awkward.util.OrderedDict()
        for n, x in fields.items():
            self._fields[n] = x

    def __getitem__(self, key):
        return self._fields[key]

    def __setitem__(self, key, value):
        if isinstance(value, Type):
            self._fields[key] = value
        else:
            self._fields[key] = self._finaltype(value)

    def __delitem__(self, key):
        del self._fields[key]

    def __and__(self, other):
        out = TableType.__new__(TableType)
        out._fields = awkward.util.OrderedDict(list(self._fields.items()) + list(other._fields.items()))
        return out

    def _subrepr(self, labeled, seen):
        return "TableType({0})".format(", ".join("{0}={1}".format(n, x._repr(labeled, seen) if isinstance(x, Type) else repr(x)) for n, x in self._fields.items()))

    def _substr(self, labeled, seen, indent):
        width = max(len(repr(n)) for n in self._fields.keys())
        subindent = indent + (" " * width) + "    "
        out = []
        for n, x in self._fields.items():
            if isinstance(x, Type):
                to = x._str(labeled, seen, subindent).lstrip(" ")
            else:
                to = str(x)
            out.append(("{0}{1:%ds} -> {2}" % width).format(indent, repr(n), to))
        return "\n".join(out).lstrip(" ")

    def _eq(self, other, seen):
        if self is other:
            return True
        elif id(self) in seen:
            return False
        else:
            seen.add(id(self))
            return isinstance(other, TableType) and [(n, self._fields[n]) for n in sorted(self._fields)] == [(n, other._fields[n]) for n in sorted(other._fields)]

    def __hash__(self):
        return hash((TableType, tuple((n, self._fields[n]) for n in sorted(self._fields))))

class UnionType(Type):
    def __init__(self, *possibilities):
        self._possibilities = []
        for x in possibilities:
            self.append(x)

    def __len__(self):
        return len(self._possibilities)

    def __getitem__(self, index):
        return self._possibilities[index]

    def __setitem__(self, index, value):
        if isinstance(value, Type):
            self._possibilities[index] = value
        else:
            self._possibilities[index] = self._finaltype(value)

    def __delitem__(self, index):
        del self._possibilities[index]

    def append(self, value):
        if isinstance(value, Type):
            self._possibilities.append(value)
        else:
            self._possibilities.append(self._finaltype(value))

    def _subrepr(self, labeled, seen):
        return "UnionType({0})".format(", ".join(x._repr(labeled, seen) if isinstance(x, Type) else repr(x) for x in self._possibilities))

    def _substr(self, labeled, seen, indent):
        subs = [x._str(labeled, seen, indent + " ") if isinstance(x, Type) else str(x) for x in self._possibilities]
        def lstrip(x):
            if x.startswith(indent + " "):
                return x[len(indent) + 1:]
            else:
                return x
        width = max(len(lstrip(y)) for x in subs for y in x.split("\n"))
        out = [x + " " * (width - len(lstrip(x.split("\n")[-1]))) for x in subs]
        return "(" + (" |\n" + indent + " ").join(out) + " )"

    def _eq(self, other, seen):
        if self is other:
            return True
        elif id(self) in seen:
            return False
        else:
            seen.add(id(self))
            return isinstance(other, UnionType) and set(self._possibilities) == set(other._possibilities)

    def __hash__(self):
        return hash((UnionType, tuple(self._possibilities)))

class OptionType(Type):
    def __init__(self, type):
        self.type = type

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        if isinstance(value, Type):
            self._type = value
        else:
            self._type = self._finaltype(self._type)

    def _subrepr(self, labeled, seen):
        return "OptionType({0})".format(self._type._repr(labeled, seen) if isinstance(self._type, Type) else repr(self._type))

    def _substr(self, labeled, seen, indent):
        if isinstance(self._type, Type):
            type = self._type._str(labeled, seen, indent + "  ").lstrip(" ")
        else:
            type = str(self._type)
        return "?({0})".format(type)

    def _eq(self, other, seen):
        if self is other:
            return True
        elif id(self) in seen:
            return False
        else:
            seen.add(id(self))
            return isinstance(other, OptionType) and self._type == other._type

    def __hash__(self):
        return hash((OptionType, self._type))



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import codecs
import collections
import numbers

import numpy

import awkward.array.base
import awkward.util
from awkward.array.chunked import PartitionedArray, AppendableArray
from awkward.array.indexed import IndexedMaskedArray, UnionArray
from awkward.array.jagged import JaggedArray
from awkward.array.masked import BitMaskedArray
from awkward.array.objects import ObjectArray
from awkward.array.table import Table

def fromiter(iterable, chunksize=1024, maskmissing=True, references=False):
    if references:
        raise NotImplementedError    # keep all ids in a hashtable to create pointers (IndexedArray)

    tobytes = lambda x: x.tobytes()
    tostring = lambda x: codecs.utf_8_decode(x.tobytes())[0]

    def insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj):
        if len(chunks) == 0 or offsets[-1] - offsets[-2] == len(chunks[-1]):
            chunks.append(newchunk(obj))
            offsets.append(offsets[-1])

        if ismine(obj, chunks[-1]):
            chunks[-1] = promote(obj, chunks[-1])
            fillobj(obj, chunks[-1], offsets[-1] - offsets[-2])
            offsets[-1] += 1

        elif isinstance(chunks[-1], IndexedMaskedArray) and len(chunks[-1]._content) == 0:
            chunks[-1]._content = newchunk(obj)

            nextindex = chunks[-1]._nextindex
            chunks[-1]._nextindex += 1
            chunks[-1]._index[offsets[-1] - offsets[-2]] = nextindex

            chunks[-1]._content = promote(obj, chunks[-1]._content)
            fillobj(obj, chunks[-1]._content, nextindex)
            offsets[-1] += 1

        elif isinstance(chunks[-1], IndexedMaskedArray) and ismine(obj, chunks[-1]._content):
            nextindex = chunks[-1]._nextindex
            chunks[-1]._nextindex += 1
            chunks[-1]._index[offsets[-1] - offsets[-2]] = nextindex

            chunks[-1]._content = promote(obj, chunks[-1]._content)
            fillobj(obj, chunks[-1]._content, nextindex)
            offsets[-1] += 1

        elif isinstance(chunks[-1], UnionArray) and any(isinstance(content, IndexedMaskedArray) and ismine(obj, content._content) for content in chunks[-1]._contents):
            for tag in range(len(chunks[-1]._contents)):
                if isinstance(chunks[-1]._contents[tag], IndexedMaskedArray) and ismine(obj, chunks[-1]._contents[tag]._content):
                    nextindex_union = chunks[-1]._nextindex[tag]
                    chunks[-1]._nextindex[tag] += 1

                    nextindex_mask = chunks[-1]._contents[tag]._nextindex
                    chunks[-1]._contents[tag]._nextindex += 1
                    chunks[-1]._contents[tag]._index[nextindex_union] = nextindex_mask

                    chunks[-1]._contents[tag]._content = promote(obj, chunks[-1]._contents[tag]._content)
                    fillobj(obj, chunks[-1]._contents[tag]._content, nextindex_mask)

                    chunks[-1]._tags[offsets[-1] - offsets[-2]] = tag
                    chunks[-1]._index[offsets[-1] - offsets[-2]] = nextindex_union

                    offsets[-1] += 1
                    break

        else:
            if not isinstance(chunks[-1], UnionArray):
                chunks[-1] = UnionArray(numpy.empty(chunksize, dtype=awkward.array.base.AwkwardArray.INDEXTYPE),
                                        numpy.empty(chunksize, dtype=awkward.array.base.AwkwardArray.INDEXTYPE),
                                        [chunks[-1]])
                chunks[-1]._nextindex = [offsets[-1] - offsets[-2]]
                chunks[-1]._tags[: offsets[-1] - offsets[-2]] = 0
                chunks[-1]._index[: offsets[-1] - offsets[-2]] = numpy.arange(offsets[-1] - offsets[-2], dtype=awkward.array.base.AwkwardArray.INDEXTYPE)
                chunks[-1]._contents = list(chunks[-1]._contents)

            if not any(ismine(obj, content) for content in chunks[-1]._contents):
                chunks[-1]._nextindex.append(0)
                chunks[-1]._contents.append(newchunk(obj))

            for tag in range(len(chunks[-1]._contents)):
                if ismine(obj, chunks[-1]._contents[tag]):
                    nextindex = chunks[-1]._nextindex[tag]
                    chunks[-1]._nextindex[tag] += 1

                    chunks[-1]._contents[tag] = promote(obj, chunks[-1]._contents[tag])
                    fillobj(obj, chunks[-1]._contents[tag], nextindex)

                    chunks[-1]._tags[offsets[-1] - offsets[-2]] = tag
                    chunks[-1]._index[offsets[-1] - offsets[-2]] = nextindex

                    offsets[-1] += 1
                    break

    def fill(obj, chunks, offsets):
        if obj is None:
            # anything with None -> IndexedMaskedArray

            if len(chunks) == 0 or offsets[-1] - offsets[-2] == len(chunks[-1]):
                chunks.append(IndexedMaskedArray(numpy.empty(chunksize, dtype=awkward.array.base.AwkwardArray.INDEXTYPE), []))
                chunks[-1]._nextindex = 0
                offsets.append(offsets[-1])

            if isinstance(chunks[-1], UnionArray) and any(isinstance(content, IndexedMaskedArray) for content in chunks[-1]._contents):
                for tag in range(len(chunks[-1]._contents)):
                    if isinstance(chunks[-1]._contents[tag], IndexedMaskedArray):
                        nextindex = chunks[-1]._nextindex[tag]
                        chunks[-1]._nextindex[tag] += 1

                        chunks[-1]._contents[tag]._index[nextindex] = chunks[-1]._contents[tag]._maskedwhen

                        chunks[-1]._tags[offsets[-1] - offsets[-2]] = tag
                        chunks[-1]._index[offsets[-1] - offsets[-2]] = nextindex

                        offsets[-1] += 1
                        break

            else:
                if not isinstance(chunks[-1], IndexedMaskedArray):
                    chunks[-1] = IndexedMaskedArray(numpy.empty(chunksize, dtype=awkward.array.base.AwkwardArray.INDEXTYPE), chunks[-1])
                    chunks[-1]._index[: offsets[-1] - offsets[-2]] = numpy.arange(offsets[-1] - offsets[-2], dtype=awkward.array.base.AwkwardArray.INDEXTYPE)
                    chunks[-1]._nextindex = offsets[-1] - offsets[-2]

                chunks[-1]._index[offsets[-1] - offsets[-2]] = chunks[-1]._maskedwhen
                offsets[-1] += 1

        elif isinstance(obj, (bool, numpy.bool, numpy.bool_)):
            # bool -> Numpy bool_

            def newchunk(obj):
                return numpy.empty(chunksize, dtype=numpy.bool_)

            def ismine(obj, x):
                return isinstance(x, numpy.ndarray) and x.dtype == numpy.dtype(numpy.bool_)

            def promote(obj, x):
                return x

            def fillobj(obj, array, where):
                array[where] = obj

            insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

        elif isinstance(obj, (numbers.Integral, numpy.integer)):
            # int -> Numpy int64, float64, or complex128 (promotes to largest)

            def newchunk(obj):
                return numpy.empty(chunksize, dtype=numpy.int64)

            def ismine(obj, x):
                return isinstance(x, numpy.ndarray) and issubclass(x.dtype.type, numpy.number)

            def promote(obj, x):
                return x

            def fillobj(obj, array, where):
                array[where] = obj

            insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

        elif isinstance(obj, (numbers.Real, numpy.floating)):
            # float -> Numpy int64, float64, or complex128 (promotes to largest)

            def newchunk(obj):
                return numpy.empty(chunksize, dtype=numpy.int64)

            def ismine(obj, x):
                return isinstance(x, numpy.ndarray) and issubclass(x.dtype.type, numpy.number)

            def promote(obj, x):
                if issubclass(x.dtype.type, numpy.floating):
                    return x
                else:
                    return x.astype(numpy.float64)

            def fillobj(obj, array, where):
                array[where] = obj

            insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

        elif isinstance(obj, (numbers.Complex, numpy.complex, numpy.complexfloating)):
            # complex -> Numpy int64, float64, or complex128 (promotes to largest)

            def newchunk(obj):
                return numpy.empty(chunksize, dtype=numpy.complex128)

            def ismine(obj, x):
                return isinstance(x, numpy.ndarray) and issubclass(x.dtype.type, numpy.number)

            def promote(obj, x):
                if issubclass(x.dtype.type, numpy.complexfloating):
                    return x
                else:
                    return x.astype(numpy.complex128)

            def fillobj(obj, array, where):
                array[where] = obj

            insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

        elif isinstance(obj, bytes):
            # bytes -> VirtualObjectArray of JaggedArray

            def newchunk(obj):
                out = VirtualObjectArray(tobytes, JaggedArray.fromoffsets(
                    numpy.zeros(chunksize + 1, dtype=awkward.array.base.AwkwardArray.INDEXTYPE),
                    AppendableArray.empty(lambda: numpy.empty(chunksize, dtype=awkward.array.base.AwkwardArray.CHARTYPE))))
                out._content._starts[0] = 0
                return out

            def ismine(obj, x):
                return isinstance(x, VirtualObjectArray) and (x._generator is tobytes or x._generator is tostring)

            def promote(obj, x):
                return x

            def fillobj(obj, array, where):
                array._content._stops[where] = array._content._starts[where] + len(obj)
                array._content._content.extend(numpy.fromstring(obj, dtype=awkward.array.base.AwkwardArray.CHARTYPE))
                
            insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

        elif isinstance(obj, awkward.util.string):
            # str -> VirtualObjectArray of JaggedArray

            def newchunk(obj):
                out = VirtualObjectArray(tostring, JaggedArray.fromoffsets(
                    numpy.zeros(chunksize + 1, dtype=awkward.array.base.AwkwardArray.INDEXTYPE),
                    AppendableArray.empty(lambda: numpy.empty(chunksize, dtype=awkward.array.base.AwkwardArray.CHARTYPE))))
                out._content._starts[0] = 0
                return out

            def ismine(obj, x):
                return isinstance(x, VirtualObjectArray) and (x._generator is tobytes or x._generator is tostring)

            def promote(obj, x):
                if x._generator is tostring:
                    return x
                else:
                    return VirtualObjectArray(tostring, x._content)

            def fillobj(obj, array, where):
                bytes = codecs.utf_8_encode(obj)[0]
                array._content._stops[where] = array._content._starts[where] + len(bytes)
                array._content._content.extend(numpy.fromstring(bytes, dtype=awkward.array.base.AwkwardArray.CHARTYPE))
                
            insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

        elif isinstance(obj, dict):
            # dict keys -> Table columns

            def newchunk(obj):
                return Table(chunksize, collections.OrderedDict((n, []) for n in obj))

            if maskmissing:
                def ismine(obj, x):
                    return isinstance(x, Table)

                def promote(obj, x):
                    for n in obj:
                        if not n in x._content:
                            x._content[n] = IndexedMaskedArray(numpy.empty(chunksize, dtype=awkward.array.base.AwkwardArray.INDEXTYPE), [])
                            x._content[n]._index[: offsets[-1] - offsets[-2]] = x._content[n]._maskedwhen
                            x._content[n]._nextindex = 0
                    return x

            else:
                def ismine(obj, x):
                    return isinstance(x, Table) and all(n in x._content for n in obj)

                def promote(obj, x):
                    return x

            def fillobj(obj, array, where):
                for n in obj:
                    if len(array._content[n]) == 0:
                        subchunks = []
                        suboffsets = [offsets[-2]]
                    else:
                        subchunks = [array._content[n]]
                        suboffsets = [offsets[-2], offsets[-1]]

                    fill(obj[n], subchunks, suboffsets)
                    array._content[n] = subchunks[-1]

            insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

        elif isinstance(obj, tuple):
            # tuple items -> Table columns

            def newchunk(obj):
                return Table(chunksize, collections.OrderedDict(("_" + str(i), []) for i in range(len(obj))))

            def ismine(obj, x):
                return isinstance(x, Table) and list(x._content) == ["_" + str(i) for i in range(len(obj))]

            def promote(obj, x):
                return x

            def fillobj(obj, array, where):
                for i, x in enumerate(obj):
                    n = "_" + str(i)
                    if len(array._content[n]) == 0:
                        subchunks = []
                        suboffsets = [offsets[-2]]
                    else:
                        subchunks = [array._content[n]]
                        suboffsets = [offsets[-2], offsets[-1]]

                    fill(x, subchunks, suboffsets)
                    array._content[n] = subchunks[-1]

            insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

        else:
            try:
                it = iter(obj)

            except TypeError:
                # object attributes -> Table columns

                def newchunk(obj):
                    return NamedTable(chunksize, obj.__class__.__name__, collections.OrderedDict((n, []) for n in dir(obj) if not n.startswith("_")))

                if maskmissing:
                    def ismine(obj, x):
                        return isinstance(x, NamedTable) and obj.__class__.__name__ == x._name

                    def promote(obj, x):
                        for n in dir(obj):
                            if not n.startswith("_") and not n in x._content:
                                x._content[n] = IndexedMaskedArray(numpy.empty(chunksize, dtype=awkward.array.base.AwkwardArray.INDEXTYPE), [])
                                x._content[n]._index[: offsets[-1] - offsets[-2]] = x._content[n]._maskedwhen
                                x._content[n]._nextindex = 0
                        return x

                else:
                    def ismine(obj, x):
                        return isinstance(x, NamedTable) and obj.__class__.__name__ == x._name and all(n in x._content for n in dir(obj) if not n.startswith("_"))

                    def promote(obj, x):
                        return x

                def fillobj(obj, array, where):
                    for n in dir(obj):
                        if not n.startswith("_"):
                            if len(array._content[n]) == 0:
                                subchunks = []
                                suboffsets = [offsets[-2]]
                            else:
                                subchunks = [array._content[n]]
                                suboffsets = [offsets[-2], offsets[-1]]

                            fill(getattr(obj, n), subchunks, suboffsets)
                            array._content[n] = subchunks[-1]

                insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

            else:
                # iterable -> JaggedArray (and recurse)

                def newchunk(obj):
                    out = JaggedArray.fromoffsets(numpy.zeros(chunksize + 1, dtype=awkward.array.base.AwkwardArray.INDEXTYPE), PartitionedArray([0], []))
                    out._starts[0] = 0
                    out._content._offsets = [0]  # as an appendable list, not a Numpy array
                    return out

                def ismine(obj, x):
                    return isinstance(x, JaggedArray)

                def promote(obj, x):
                    return x

                def fillobj(obj, array, where):
                    array._stops[where] = array._starts[where]
                    for x in it:
                        fill(x, array._content._chunks, array._content._offsets)
                        array._stops[where] += 1

                insert(obj, chunks, offsets, newchunk, ismine, promote, fillobj)

    def trim(length, array):
        if isinstance(array, numpy.ndarray):
            if len(array) == length:
                return array                          # the length is right: don't copy it
            else:
                return numpy.array(array[:length])    # copy so that the base can be deleted

        elif isinstance(array, PartitionedArray):
            for i in range(len(array._chunks)):
                array._chunks[i] = trim(array._offsets[i + 1] - array._offsets[i], array._chunks[i])
            return array

        elif isinstance(array, IndexedMaskedArray):
            index = trim(length, array._index)
            selection = (index != array._maskedwhen)
            content = trim(index[selection][-1] + 1, array._content)

            if isinstance(content, numpy.ndarray):
                # for simple types, IndexedMaskedArray wastes space; convert to an Arrow-like BitMaskedArray
                mask = numpy.zeros(length, dtype=awkward.array.base.AwkwardArray.MASKTYPE)
                mask[selection] = True

                newcontent = numpy.empty(length, dtype=content.dtype)
                newcontent[selection] = content

                return BitMaskedArray.fromboolmask(mask, newcontent, maskedwhen=False, lsb=True)

            else:
                # for complex types, IndexedMaskedArray saves space; keep it
                return IndexedMaskedArray(index, content)

        elif isinstance(array, UnionArray):
            tags = trim(length, array._tags)
            index = trim(length, array._index)

            contents = []
            for tag, content in enumerate(array._contents):
                length = index[tags == tag][-1] + 1
                contents.append(trim(length, content))

            return UnionArray(tags, index, contents)

        elif isinstance(array, JaggedArray):
            offsets = array.offsets                   # fill creates aliased starts/stops
            if len(offsets) != length + 1:
                offsets = numpy.array(offsets[: length + 1])

            return JaggedArray.fromoffsets(offsets, trim(offsets[-1], array._content))

        elif isinstance(array, NamedTable):
            return NamedTable(length, array._name, collections.OrderedDict((n, trim(length, x)) for n, x in array._content.items()))

        elif isinstance(array, Table):
            return Table(length, collections.OrderedDict((n, trim(length, x)) for n, x in array._content.items()))

        elif isinstance(array, VirtualObjectArray):
            return VirtualObjectArray(array._generator, trim(length, array._content))

        else:
            raise AssertionError(array)

    chunks = []
    offsets = [0]
    length = 0
    for x in iterable:
        fill(x, chunks, offsets)
        length += 1
        
    return trim(length, PartitionedArray(offsets, chunks))



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

import numpy

from awkward import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    ################### old tests

    # def test_masked_get(self):
    #     a = MaskedArray([True, False, True, False, True, False, True, False, True, False], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], maskedwhen=True)
    #     self.assertEqual(a.tolist(), [None, 1.1, None, 3.3, None, 5.5, None, 7.7, None, 9.9])
    #     self.assertTrue(numpy.ma.is_masked(a[0]))
    #     self.assertFalse(numpy.ma.is_masked(a[1]))
    #     self.assertEqual(a[5:].tolist(), [5.5, None, 7.7, None, 9.9])
    #     self.assertFalse(numpy.ma.is_masked(a[5:][0]))
    #     self.assertTrue(numpy.ma.is_masked(a[5:][1]))
    #     self.assertEqual(a[[3, 2, 1]].tolist(), [3.3, None, 1.1])
    #     self.assertEqual(a[[True, True, True, True, True, False, False, False, False, False]].tolist(), [None, 1.1, None, 3.3, None])

    # def test_masked_get_flip(self):
    #     a = MaskedArray([False, True, False, True, False, True, False, True, False, True], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], maskedwhen=False)
    #     self.assertEqual(a.tolist(), [None, 1.1, None, 3.3, None, 5.5, None, 7.7, None, 9.9])
    #     self.assertTrue(numpy.ma.is_masked(a[0]))
    #     self.assertFalse(numpy.ma.is_masked(a[1]))
    #     self.assertEqual(a[5:].tolist(), [5.5, None, 7.7, None, 9.9])
    #     self.assertFalse(numpy.ma.is_masked(a[5:][0]))
    #     self.assertTrue(numpy.ma.is_masked(a[5:][1]))
    #     self.assertEqual(a[[3, 2, 1]].tolist(), [3.3, None, 1.1])
    #     self.assertEqual(a[[True, True, True, True, True, False, False, False, False, False]].tolist(), [None, 1.1, None, 3.3, None])

    # def test_bitmasked_get(self):
    #     a = BitMaskedArray.fromboolmask([True, False, True, False, True, False, True, False, True, False], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], maskedwhen=True, lsb=True)
    #     self.assertEqual(a.tolist(), [None, 1.1, None, 3.3, None, 5.5, None, 7.7, None, 9.9])
    #     self.assertTrue(numpy.ma.is_masked(a[0]))
    #     self.assertFalse(numpy.ma.is_masked(a[1]))
    #     self.assertEqual(a[5:].tolist(), [5.5, None, 7.7, None, 9.9])
    #     self.assertFalse(numpy.ma.is_masked(a[5:][0]))
    #     self.assertTrue(numpy.ma.is_masked(a[5:][1]))
    #     self.assertEqual(a[[3, 2, 1]].tolist(), [3.3, None, 1.1])
    #     self.assertEqual(a[[True, True, True, True, True, False, False, False, False, False]].tolist(), [None, 1.1, None, 3.3, None])

    #     a = BitMaskedArray.fromboolmask([True, False, True, False, True, False, True, False, True, False], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], maskedwhen=True, lsb=False)
    #     self.assertEqual(a.tolist(), [None, 1.1, None, 3.3, None, 5.5, None, 7.7, None, 9.9])
    #     self.assertTrue(numpy.ma.is_masked(a[0]))
    #     self.assertFalse(numpy.ma.is_masked(a[1]))
    #     self.assertEqual(a[5:].tolist(), [5.5, None, 7.7, None, 9.9])
    #     self.assertFalse(numpy.ma.is_masked(a[5:][0]))
    #     self.assertTrue(numpy.ma.is_masked(a[5:][1]))
    #     self.assertEqual(a[[3, 2, 1]].tolist(), [3.3, None, 1.1])
    #     self.assertEqual(a[[True, True, True, True, True, False, False, False, False, False]].tolist(), [None, 1.1, None, 3.3, None])

    # def test_bitmasked_get_flip(self):
    #     a = BitMaskedArray.fromboolmask([False, True, False, True, False, True, False, True, False, True], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], maskedwhen=False, lsb=True)
    #     self.assertEqual(a.tolist(), [None, 1.1, None, 3.3, None, 5.5, None, 7.7, None, 9.9])
    #     self.assertTrue(numpy.ma.is_masked(a[0]))
    #     self.assertFalse(numpy.ma.is_masked(a[1]))
    #     self.assertEqual(a[5:].tolist(), [5.5, None, 7.7, None, 9.9])
    #     self.assertFalse(numpy.ma.is_masked(a[5:][0]))
    #     self.assertTrue(numpy.ma.is_masked(a[5:][1]))
    #     self.assertEqual(a[[3, 2, 1]].tolist(), [3.3, None, 1.1])
    #     self.assertEqual(a[[True, True, True, True, True, False, False, False, False, False]].tolist(), [None, 1.1, None, 3.3, None])

    #     a = BitMaskedArray.fromboolmask([False, True, False, True, False, True, False, True, False, True], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], maskedwhen=False, lsb=False)
    #     self.assertEqual(a.tolist(), [None, 1.1, None, 3.3, None, 5.5, None, 7.7, None, 9.9])
    #     self.assertTrue(numpy.ma.is_masked(a[0]))
    #     self.assertFalse(numpy.ma.is_masked(a[1]))
    #     self.assertEqual(a[5:].tolist(), [5.5, None, 7.7, None, 9.9])
    #     self.assertFalse(numpy.ma.is_masked(a[5:][0]))
    #     self.assertTrue(numpy.ma.is_masked(a[5:][1]))
    #     self.assertEqual(a[[3, 2, 1]].tolist(), [3.3, None, 1.1])
    #     self.assertEqual(a[[True, True, True, True, True, False, False, False, False, False]].tolist(), [None, 1.1, None, 3.3, None])

    # def test_bitmasked_arrow(self):
    #     # Apache Arrow layout example
    #     # https://github.com/apache/arrow/blob/master/format/Layout.md#null-bitmaps
    #     a = BitMaskedArray.fromboolmask([True, True, False, True, False, True], [0, 1, 999, 2, 999, 3], maskedwhen=False, lsb=True)
    #     self.assertEqual(a.tolist(), [0, 1, None, 2, None, 3])

    #     # extra gunk at the end of the array
    #     a = BitMaskedArray.fromboolmask([True, True, False, True, False, True, True, True], [0, 1, 999, 2, 999, 3], maskedwhen=False, lsb=True)
    #     self.assertEqual(a.tolist(), [0, 1, None, 2, None, 3])

    #     # opposite sign
    #     a = BitMaskedArray.fromboolmask([True, True, False, True, False, True, False, False], [0, 1, 999, 2, 999, 3], maskedwhen=False, lsb=True)
    #     self.assertEqual(a.tolist(), [0, 1, None, 2, None, 3])

    #     # doubled
    #     a = BitMaskedArray.fromboolmask([True, True, False, True, False, True, True, True, False, True, False, True], [0, 1, 999, 2, 999, 3, 0, 1, 999, 2, 999, 3], maskedwhen=False, lsb=True)
    #     self.assertEqual(a.tolist(), [0, 1, None, 2, None, 3, 0, 1, None, 2, None, 3])



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import collections
import numbers

import numpy

import awkward.array.base

class MaskedArray(awkward.array.base.AwkwardArray):
    def __init__(self, mask, content, maskedwhen=True):
        raise NotImplementedError

# class MaskedArray(awkward.array.base.AwkwardArray):
#     def __init__(self, mask, content, maskedwhen=True):
#         self.mask = mask
#         self.content = content
#         self.maskedwhen = maskedwhen

#     @property
#     def mask(self):
#         return self._mask

#     @mask.setter
#     def mask(self, value):
#         value = self._toarray(value, self.MASKTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray))

#         if len(value.shape) != 1:
#             raise TypeError("mask must have 1-dimensional shape")
#         if value.shape[0] == 0:
#             value = value.view(self.MASKTYPE)
#         if not issubclass(value.dtype.type, (numpy.bool_, numpy.bool)):
#             raise TypeError("mask must have boolean dtype")

#         self._mask = value

#     @property
#     def boolmask(self):
#         return self._mask

#     @boolmask.setter
#     def boolmask(self, value):
#         self.mask = value

#     @property
#     def content(self):
#         return self._content

#     @content.setter
#     def content(self, value):
#         self._content = self._toarray(value, self.CHARTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray))
        
#     @property
#     def maskedwhen(self):
#         return self._maskedwhen

#     @maskedwhen.setter
#     def maskedwhen(self, value):
#         self._maskedwhen = bool(value)

#     @property
#     def dtype(self):
#         return self._content.dtype

#     @property
#     def shape(self):
#         return self._content.shape

#     def __len__(self):
#         return len(self._content)

#     def __getitem__(self, where):
#         if self._isstring(where):
#             return MaskedArray(self._mask, self._content[where], maskedwhen=self._maskedwhen)

#         if not isinstance(where, tuple):
#             where = (where,)
#         head, tail = where[0], where[1:]

#         if isinstance(head, (numbers.Integral, numpy.integer)):
#             if self._mask[head] == self._maskedwhen:
#                 return numpy.ma.masked
#             else:
#                 return self._content[self._singleton(where)]
#         else:
#             return MaskedArray(self._mask[head], self._content[self._singleton(where)], maskedwhen=self._maskedwhen)

class BitMaskedArray(MaskedArray):
    def __init__(self, mask, content, maskedwhen=True, lsb=True):
        raise NotImplementedError

# class BitMaskedArray(MaskedArray):
#     @staticmethod
#     def fromboolmask(mask, content, maskedwhen=True, lsb=True):
#         out = BitMaskedArray([], content, maskedwhen=maskedwhen, lsb=lsb)
#         out.boolmask = mask
#         return out

#     def __init__(self, mask, content, maskedwhen=True, lsb=True):
#         self.mask = mask
#         self.content = content
#         self.maskedwhen = maskedwhen
#         self.lsb = lsb

#     @property
#     def mask(self):
#         return self._mask

#     @mask.setter
#     def mask(self, value):
#         value = self._toarray(value, self.BITMASKTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray))

#         if len(value.shape) != 1:
#             raise TypeError("mask must have 1-dimensional shape")

#         self._mask = value.view(self.BITMASKTYPE)

#     @property
#     def boolmask(self):
#         out = numpy.unpackbits(self._mask)
#         if self._lsb:
#             out = out.reshape(-1, 8)[:,::-1].reshape(-1)
#         return out.view(self.MASKTYPE)[:len(self._content)]

#     @boolmask.setter
#     def boolmask(self, value):
#         value = numpy.array(value, copy=False)

#         if len(value.shape) != 1:
#             raise TypeError("boolmask must have 1-dimensional shape")
#         if not issubclass(value.dtype.type, (numpy.bool, numpy.bool_)):
#             raise TypeError("boolmask must have boolean type")

#         if self._lsb:
#             # maybe pad the length for reshape
#             length = 8*((len(value) + 8 - 1) >> 3)   # ceil(len(value) / 8.0) * 8
#             if length != len(value):
#                 out = numpy.empty(length, dtype=numpy.bool_)
#                 out[:len(value)] = value
#             else:
#                 out = value

#             # reverse the order in groups of 8
#             out = out.reshape(-1, 8)[:,::-1].reshape(-1)

#         else:
#             # numpy.packbits encodes as msb (most significant bit); already in the right order
#             out = value

#         self._mask = numpy.packbits(out)
        
#     @property
#     def lsb(self):
#         return self._lsb

#     @lsb.setter
#     def lsb(self, value):
#         self._lsb = bool(value)

#     def _maskat(self, where):
#         bytepos = numpy.right_shift(where, 3)    # where // 8
#         bitpos  = where - 8*bytepos              # where % 8

#         if self.lsb:
#             bitmask = numpy.left_shift(1, bitpos)
#         else:
#             bitmask = numpy.right_shift(128, bitpos)

#         if isinstance(bitmask, numpy.ndarray):
#             bitmask = bitmask.astype(self.BITMASKTYPE)
#         else:
#             bitmask = self.BITMASKTYPE.type(bitmask)

#         return bytepos, bitmask

#     def _maskwhere(self, where):
#         if isinstance(where, (numbers.Integral, numpy.integer)):
#             bytepos, bitmask = self._maskat(where)
#             return numpy.bitwise_and(self._mask[bytepos], bitmask) != 0

#         elif isinstance(where, slice):
#             # assumes a small slice; for a big slice, it could be faster to unpack the whole mask
#             return self._maskwhere(numpy.arange(*where.indices(len(self._content))))

#         else:
#             where = numpy.array(where, copy=False)
#             if len(where.shape) == 1 and issubclass(where.dtype.type, numpy.integer):
#                 byteposes, bitmasks = self._maskat(where)
#                 numpy.bitwise_and(bitmasks, self._mask[byteposes], bitmasks)
#                 return bitmasks.astype(numpy.bool_)
        
#             elif len(where.shape) == 1 and issubclass(where.dtype.type, (numpy.bool, numpy.bool_)):
#                 # scales with the size of the mask anyway, so go ahead and unpack the whole mask
#                 unpacked = numpy.unpackbits(self._mask).view(self.MASKTYPE)

#                 if self.lsb:
#                     unpacked = unpacked.reshape(-1, 8)[:,::-1].reshape(-1)[:len(where)]
#                 else:
#                     unpacked = unpacked[:len(where)]

#                 return unpacked[where]

#             else:
#                 raise TypeError("cannot interpret shape {0}, dtype {1} as a fancy index or mask".format(where.shape, where.dtype))

#     def _setmask(self, where, valid):
#         if isinstance(where, (numbers.Integral, numpy.integer)):        
#             bytepos, bitmask = self._maskat(where)
#             if self._maskedwhen != valid:
#                 self._mask[bytepos] |= bitmask
#             else:
#                 self._mask[bytepos] &= numpy.bitwise_not(bitmask)

#         elif isinstance(where, slice):
#             # assumes a small slice; for a big slice, it could be faster to unpack the whole mask
#             self._setmask(numpy.arange(*where.indices(len(self._content))), valid)

#         else:
#             where = numpy.array(where, copy=False)
#             if len(where.shape) == 1 and issubclass(where.dtype.type, numpy.integer):
#                 bytepos, bitmask = self._maskat(where)
#                 if self._maskedwhen != valid:
#                     numpy.bitwise_or.at(self._mask, bytepos, bitmask)
#                 else:
#                     numpy.bitwise_and.at(self._mask, bytepos, numpy.bitwise_not(bitmask))

#             elif len(where.shape) == 1 and issubclass(where.dtype.type, (numpy.bool, numpy.bool_)):
#                 tmp = self.boolmask
#                 if self._maskedwhen != valid:
#                     tmp[where] = True
#                 else:
#                     tmp[where] = False
#                 self.boolmask = tmp

#             else:
#                 raise TypeError("cannot interpret shape {0}, dtype {1} as a fancy index or mask".format(where.shape, where.dtype))

#     def __getitem__(self, where):
#         if self._isstring(where):
#             return MaskedArray(self._mask, self._content[where], maskedwhen=self._maskedwhen)

#         if not isinstance(where, tuple):
#             where = (where,)
#         head, tail = where[0], where[1:]

#         if isinstance(head, (numbers.Integral, numpy.integer)):
#             if self._maskwhere(head) == self._maskedwhen:
#                 return numpy.ma.masked
#             else:
#                 return self._content[self._singleton(where)]

#         else:
#             return MaskedArray(self._maskwhere(head), self._content[self._singleton(where)], maskedwhen=self._maskedwhen)



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

import numpy

from awkward import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    ################### old tests

#     def test_chunked_iteration(self):
#         a = ChunkedArray([[], [0, 1, 2, 3, 4], [5, 6], [], [7, 8, 9], []])
#         self.assertEqual(a.tolist(), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

#         a = ChunkedArray([[0, 1, 2, 3, 4], [5, 6], [], [7, 8, 9], []])
#         self.assertEqual(a.tolist(), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

#         a = ChunkedArray([[], [0, 1, 2, 3, 4], [5, 6], [], [7, 8, 9]])
#         self.assertEqual(a.tolist(), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

#         a = ChunkedArray([[]])
#         self.assertEqual(a.tolist(), [])

#         a = ChunkedArray([])
#         self.assertEqual(a.tolist(), [])

#     def test_chunked_dtype(self):
#         a = ChunkedArray([[], [0, 1, 2, 3, 4], [5, 6], [], [7, 8, 9], []])
#         self.assertEqual(a.dtype, numpy.dtype(int))

#         a = ChunkedArray([[0, 1, 2, 3, 4], [5, 6], [], [7, 8, 9], []])
#         self.assertEqual(a.dtype, numpy.dtype(int))

#         a = ChunkedArray([])
#         self.assertRaises(ValueError, lambda: a.dtype)

#     def test_chunked_get(self):
#         a = ChunkedArray([[], [0, 1, 2, 3, 4], [5, 6], [], [7, 8, 9], []])
#         self.assertEqual([a[i] for i in range(10)], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])

#         self.assertEqual([a[i : i + 1].tolist() for i in range(10)], [[0], [1], [2], [3], [4], [5], [6], [7], [8], [9]])
#         self.assertEqual([a[i : i + 2].tolist() for i in range(10 - 1)], [[0, 1], [1, 2], [2, 3], [3, 4], [4, 5], [5, 6], [6, 7], [7, 8], [8, 9]])
#         self.assertEqual([a[i : i + 3].tolist() for i in range(10 - 2)], [[0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, 8], [7, 8, 9]])
#         self.assertEqual([a[i : i + 4].tolist() for i in range(10 - 3)], [[0, 1, 2, 3], [1, 2, 3, 4], [2, 3, 4, 5], [3, 4, 5, 6], [4, 5, 6, 7], [5, 6, 7, 8], [6, 7, 8, 9]])
#         self.assertEqual([a[i : i + 4 : 2].tolist() for i in range(10 - 3)], [[0, 2], [1, 3], [2, 4], [3, 5], [4, 6], [5, 7], [6, 8]])
#         self.assertEqual([a[i : i + 4 : 3].tolist() for i in range(10 - 3)], [[0, 3], [1, 4], [2, 5], [3, 6], [4, 7], [5, 8], [6, 9]])
#         self.assertEqual([a[i + 4 : i : -1].tolist() for i in range(10 - 4)], [[4, 3, 2, 1], [5, 4, 3, 2], [6, 5, 4, 3], [7, 6, 5, 4], [8, 7, 6, 5], [9, 8, 7, 6]])
#         self.assertEqual([a[i + 4 : i : -2].tolist() for i in range(10 - 4)], [[4, 2], [5, 3], [6, 4], [7, 5], [8, 6], [9, 7]])
#         self.assertEqual([a[i + 4 : i : -3].tolist() for i in range(10 - 4)], [[4, 1], [5, 2], [6, 3], [7, 4], [8, 5], [9, 6]])

#         self.assertEqual(a[4:].tolist(), [4, 5, 6, 7, 8, 9])
#         self.assertEqual(a[5:].tolist(), [5, 6, 7, 8, 9])
#         self.assertEqual(a[6:].tolist(), [6, 7, 8, 9])
#         self.assertEqual(a[7:].tolist(), [7, 8, 9])
#         self.assertEqual(a[8:].tolist(), [8, 9])
#         self.assertEqual(a[:4].tolist(), [0, 1, 2, 3])
#         self.assertEqual(a[:5].tolist(), [0, 1, 2, 3, 4])
#         self.assertEqual(a[:6].tolist(), [0, 1, 2, 3, 4, 5])
#         self.assertEqual(a[:7].tolist(), [0, 1, 2, 3, 4, 5, 6])
#         self.assertEqual(a[:8].tolist(), [0, 1, 2, 3, 4, 5, 6, 7])

#         self.assertEqual(a[4::2].tolist(), [4, 6, 8])
#         self.assertEqual(a[5::2].tolist(), [5, 7, 9])
#         self.assertEqual(a[6::2].tolist(), [6, 8])
#         self.assertEqual(a[7::2].tolist(), [7, 9])
#         self.assertEqual(a[8::2].tolist(), [8])
#         self.assertEqual(a[:4:2].tolist(), [0, 2])
#         self.assertEqual(a[:5:2].tolist(), [0, 2, 4])
#         self.assertEqual(a[:6:2].tolist(), [0, 2, 4])
#         self.assertEqual(a[:7:2].tolist(), [0, 2, 4, 6])
#         self.assertEqual(a[:8:2].tolist(), [0, 2, 4, 6])

#         self.assertEqual(a[4::-1].tolist(), [4, 3, 2, 1, 0])
#         self.assertEqual(a[5::-1].tolist(), [5, 4, 3, 2, 1, 0])
#         self.assertEqual(a[6::-1].tolist(), [6, 5, 4, 3, 2, 1, 0])
#         self.assertEqual(a[7::-1].tolist(), [7, 6, 5, 4, 3, 2, 1, 0])
#         self.assertEqual(a[8::-1].tolist(), [8, 7, 6, 5, 4, 3, 2, 1, 0])
#         self.assertEqual(a[:4:-1].tolist(), [9, 8, 7, 6, 5])
#         self.assertEqual(a[:5:-1].tolist(), [9, 8, 7, 6])
#         self.assertEqual(a[:6:-1].tolist(), [9, 8, 7])
#         self.assertEqual(a[:7:-1].tolist(), [9, 8])
#         self.assertEqual(a[:8:-1].tolist(), [9])

#         self.assertEqual(a[4::-2].tolist(), [4, 2, 0])
#         self.assertEqual(a[5::-2].tolist(), [5, 3, 1])
#         self.assertEqual(a[6::-2].tolist(), [6, 4, 2, 0])
#         self.assertEqual(a[7::-2].tolist(), [7, 5, 3, 1])
#         self.assertEqual(a[8::-2].tolist(), [8, 6, 4, 2, 0])
#         self.assertEqual(a[:4:-2].tolist(), [9, 7, 5])
#         self.assertEqual(a[:5:-2].tolist(), [9, 7])
#         self.assertEqual(a[:6:-2].tolist(), [9, 7])
#         self.assertEqual(a[:7:-2].tolist(), [9])
#         self.assertEqual(a[:8:-2].tolist(), [9])

#         self.assertEqual(a[[8, 6, 4, 5, 0]].tolist(), [8, 6, 4, 5, 0])
#         self.assertEqual(a[[6, 4, 5, 0]].tolist(), [6, 4, 5, 0])
#         self.assertEqual(a[[5, 6, 4, 5, 5, 5, 0]].tolist(), [5, 6, 4, 5, 5, 5, 0])
#         self.assertRaises(IndexError, lambda: a[[8, 6, 4, 5, 0, 99]])

#         self.assertEqual(a[[True, False, True, False, True, False, True, False, True, False]].tolist(), [0, 2, 4, 6, 8])
#         self.assertEqual(a[[False, False, False, False, False, False, False, False, False, False]].tolist(), [])
#         self.assertEqual(a[[True, True, True, True, True, True, True, True, True, True]].tolist(), [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
#         self.assertRaises(IndexError, lambda: a[[True, True, True, True, True, True, True, True, True, True, True]])
#         self.assertRaises(IndexError, lambda: a[[True, True, True, True, True, True, True, True, True]])

#     def test_chunked_get2d(self):
#         a = ChunkedArray([[], [[0, 0], [1, 1], [2, 2], [3, 3], [4, 4]], [[5, 5], [6, 6]], [], [[7, 7], [8, 8], [9, 9]], []])
#         self.assertEqual([a[i].tolist() for i in range(10)], [[0, 0], [1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6], [7, 7], [8, 8], [9, 9]])
#         self.assertEqual(a[4:].tolist(), [[4, 4], [5, 5], [6, 6], [7, 7], [8, 8], [9, 9]])
#         self.assertEqual(a[[8, 6, 4, 5, 0]].tolist(), [[8, 8], [6, 6], [4, 4], [5, 5], [0, 0]])
#         self.assertEqual(a[[True, False, True, False, True, False, True, False, True, False]].tolist(), [[0, 0], [2, 2], [4, 4], [6, 6], [8, 8]])

#         a = ChunkedArray([[], [[0.0, 0.0], [1.0, 1.1], [2.0, 2.2], [3.0, 3.3], [4.0, 4.4]], [[5.0, 5.5], [6.0, 6.6]], [], [[7.0, 7.7], [8.0, 8.8], [9.0, 9.9]], []])
#         self.assertEqual([a[i, 0].tolist() for i in range(10)], [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
#         self.assertEqual([a[i, 1].tolist() for i in range(10)], [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
#         self.assertEqual(a[4:, 0].tolist(), [4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
#         self.assertEqual(a[4:, 1].tolist(), [4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
#         self.assertEqual(a[[8, 6, 4, 5, 0], 0].tolist(), [8.0, 6.0, 4.0, 5.0, 0.0])
#         self.assertEqual(a[[8, 6, 4, 5, 0], 1].tolist(), [8.8, 6.6, 4.4, 5.5, 0.0])
#         self.assertEqual(a[[True, False, True, False, True, False, True, False, True, False], 0].tolist(), [0.0, 2.0, 4.0, 6.0, 8.0])
#         self.assertEqual(a[[True, False, True, False, True, False, True, False, True, False], 1].tolist(), [0.0, 2.2, 4.4, 6.6, 8.8])

#     def test_partitioned(self):
#         a = PartitionedArray([0, 3, 3, 5], [[0, 1, 2], [], [3, 4]])
#         self.assertEqual(a.tolist(), [0, 1, 2, 3, 4])

#         a = ChunkedArray([[0, 1, 2], [], [3, 4]]).topartitioned()
#         self.assertEqual(a.tolist(), [0, 1, 2, 3, 4])
#         self.assertEqual(a.offsets.tolist(), [0, 3, 3, 5])

#     def test_partitioned_get(self):
#         a = ChunkedArray([[], [0, 1, 2, 3, 4], [5, 6], [], [7, 8, 9], []]).topartitioned()
#         self.assertEqual([a[-i] for i in range(1, 11)], [9, 8, 7, 6, 5, 4, 3, 2, 1, 0])
#         self.assertEqual([a[-i : -i - 1 : -1].tolist() for i in range(1, 11)], [[9], [8], [7], [6], [5], [4], [3], [2], [1], [0]])
#         self.assertEqual([a[-i : -i - 2 : -1].tolist() for i in range(1, 10)], [[9, 8], [8, 7], [7, 6], [6, 5], [5, 4], [4, 3], [3, 2], [2, 1], [1, 0]])
#         self.assertEqual([a[-i : -i + 1].tolist() for i in range(2, 11)], [[8], [7], [6], [5], [4], [3], [2], [1], [0]])
#         self.assertEqual([a[-i : -i + 2].tolist() for i in range(3, 11)], [[7, 8], [6, 7], [5, 6], [4, 5], [3, 4], [2, 3], [1, 2], [0, 1]])
#         self.assertEqual(a[[-2, -4, 7, 5, 3, -6, 5]].tolist(), [8, 6, 7, 5, 3, 4, 5])
#         self.assertEqual(a[[-2, -4, 7, 5, 3, -6, -5]].tolist(), [8, 6, 7, 5, 3, 4, 5])

#     def test_appendable_append(self):
#         a = AppendableArray.empty(lambda: numpy.empty(3, numpy.float64))
#         self.assertEqual(a.tolist(), [])
#         self.assertEqual(len(a.chunks), 0)
#         self.assertEqual(a.offsets, [0])

#         a.append(0.0)
#         self.assertEqual(a.tolist(), [0.0])
#         self.assertEqual(len(a.chunks), 1)
#         self.assertEqual(a.offsets, [0, 1])

#         a.append(1.1)
#         self.assertEqual(a.tolist(), [0.0, 1.1])
#         self.assertEqual(len(a.chunks), 1)
#         self.assertEqual(a.offsets, [0, 2])

#         a.append(2.2)
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2])
#         self.assertEqual(len(a.chunks), 1)
#         self.assertEqual(a.offsets, [0, 3])

#         a.append(3.3)
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3])
#         self.assertEqual(len(a.chunks), 2)
#         self.assertEqual(a.offsets, [0, 3, 4])

#         a.append(4.4)
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4])
#         self.assertEqual(len(a.chunks), 2)
#         self.assertEqual(a.offsets, [0, 3, 5])

#         a.append(5.5)
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
#         self.assertEqual(len(a.chunks), 2)
#         self.assertEqual(a.offsets, [0, 3, 6])

#         a.append(6.6)
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
#         self.assertEqual(len(a.chunks), 3)
#         self.assertEqual(a.offsets, [0, 3, 6, 7])

#         a.append(7.7)
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7])
#         self.assertEqual(len(a.chunks), 3)
#         self.assertEqual(a.offsets, [0, 3, 6, 8])

#         a.append(8.8)
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8])
#         self.assertEqual(len(a.chunks), 3)
#         self.assertEqual(a.offsets, [0, 3, 6, 9])

#         a.append(9.9)
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
#         self.assertEqual(len(a.chunks), 4)
#         self.assertEqual(a.offsets, [0, 3, 6, 9, 10])

#     def test_appendable_extend(self):
#         a = AppendableArray.empty(lambda: numpy.empty(3, numpy.float64))
#         self.assertEqual(a.tolist(), [])
#         self.assertEqual(len(a.chunks), 0)
#         self.assertEqual(a.offsets, [0])

#         a.extend([0.0, 1.1])
#         self.assertEqual(a.tolist(), [0.0, 1.1])
#         self.assertEqual(len(a.chunks), 1)
#         self.assertEqual(a.offsets, [0, 2])

#         a.extend([2.2, 3.3])
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3])
#         self.assertEqual(len(a.chunks), 2)
#         self.assertEqual(a.offsets, [0, 3, 4])

#         a.extend([4.4, 5.5])
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
#         self.assertEqual(len(a.chunks), 2)
#         self.assertEqual(a.offsets, [0, 3, 6])

#         a.extend([6.6, 7.7])
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7])
#         self.assertEqual(len(a.chunks), 3)
#         self.assertEqual(a.offsets, [0, 3, 6, 8])

#         a.extend([8.8, 9.9])
#         self.assertEqual(a.tolist(), [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
#         self.assertEqual(len(a.chunks), 4)
#         self.assertEqual(a.offsets, [0, 3, 6, 9, 10])



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest

import numpy

from awkward import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    ################### old tests

    # def test_indexed_get(self):
    #     a = IndexedArray([3, 2, 4, 2, 2, 4, 0], [0.0, 1.1, 2.2, 3.3, 4.4])
    #     self.assertEqual([x for x in a], [3.3, 2.2, 4.4, 2.2, 2.2, 4.4, 0.0])
    #     self.assertEqual([a[i] for i in range(len(a))], [3.3, 2.2, 4.4, 2.2, 2.2, 4.4, 0.0])
    #     self.assertEqual([a[i : i + 1].tolist() for i in range(len(a))], [[3.3], [2.2], [4.4], [2.2], [2.2], [4.4], [0.0]])
    #     self.assertEqual([a[i : i + 2].tolist() for i in range(len(a) - 1)], [[3.3, 2.2], [2.2, 4.4], [4.4, 2.2], [2.2, 2.2], [2.2, 4.4], [4.4, 0.0]])
    #     self.assertEqual(a[:].tolist(), [3.3, 2.2, 4.4, 2.2, 2.2, 4.4, 0.0])
    #     self.assertEqual(a[[6, 5, 4, 3, 2, 1, 0]].tolist(), [0.0, 4.4, 2.2, 2.2, 4.4, 2.2, 3.3])
    #     self.assertEqual(a[[True, False, True, False, True, False, True]].tolist(), [3.3, 4.4, 2.2, 0.0])
    #     self.assertEqual(a[[-1, -2, -3, -4, -5, -6, -7]].tolist(), [0.0, 4.4, 2.2, 2.2, 4.4, 2.2, 3.3])

    # def test_indexed_get2d(self):
    #     a = IndexedArray([3, 2, 4, 2, 2, 4, 0], [[0.0, 0.0], [1.1, 1.1], [2.2, 2.2], [3.3, 3.3], [4.4, 4.4]])
    #     self.assertEqual([x.tolist() for x in a], [[3.3, 3.3], [2.2, 2.2], [4.4, 4.4], [2.2, 2.2], [2.2, 2.2], [4.4, 4.4], [0.0, 0.0]])
    #     self.assertEqual([a[i].tolist() for i in range(len(a))], [[3.3, 3.3], [2.2, 2.2], [4.4, 4.4], [2.2, 2.2], [2.2, 2.2], [4.4, 4.4], [0.0, 0.0]])
    #     self.assertEqual(a[:].tolist(), [[3.3, 3.3], [2.2, 2.2], [4.4, 4.4], [2.2, 2.2], [2.2, 2.2], [4.4, 4.4], [0.0, 0.0]])
    #     self.assertEqual(a[[6, 5, 4, 3, 2, 1, 0]].tolist(), [[0.0, 0.0], [4.4, 4.4], [2.2, 2.2], [2.2, 2.2], [4.4, 4.4], [2.2, 2.2], [3.3, 3.3]])
    #     self.assertEqual(a[[True, False, True, False, True, False, True]].tolist(), [[3.3, 3.3], [4.4, 4.4], [2.2, 2.2], [0.0, 0.0]])

    # def test_indexed_getstruct(self):
    #     a = IndexedArray([3, 2, 4, 2, 2, 4, 0], numpy.array([(0.0, 0.0), (1.1, 1.1), (2.2, 2.2), (3.3, 3.3), (4.4, 4.4)], dtype=[("a", float), ("b", float)]))
    #     self.assertEqual([x.tolist() for x in a], [(3.3, 3.3), (2.2, 2.2), (4.4, 4.4), (2.2, 2.2), (2.2, 2.2), (4.4, 4.4), (0.0, 0.0)])
    #     self.assertEqual([a[i].tolist() for i in range(len(a))], [(3.3, 3.3), (2.2, 2.2), (4.4, 4.4), (2.2, 2.2), (2.2, 2.2), (4.4, 4.4), (0.0, 0.0)])
    #     self.assertEqual(a[:].tolist(), [(3.3, 3.3), (2.2, 2.2), (4.4, 4.4), (2.2, 2.2), (2.2, 2.2), (4.4, 4.4), (0.0, 0.0)])
    #     self.assertEqual(a[[6, 5, 4, 3, 2, 1, 0]].tolist(), [(0.0, 0.0), (4.4, 4.4), (2.2, 2.2), (2.2, 2.2), (4.4, 4.4), (2.2, 2.2), (3.3, 3.3)])
    #     self.assertEqual(a[[True, False, True, False, True, False, True]].tolist(), [(3.3, 3.3), (4.4, 4.4), (2.2, 2.2), (0.0, 0.0)])

    # def test_indexed_getempty(self):
    #     a = IndexedArray([], [0.0, 1.1, 2.2, 3.3, 4.4])
    #     self.assertEqual(a[:].tolist(), [])

    # def test_indexed_indexed(self):
    #     a = IndexedArray([6, 5, 4, 3, 2, 1, 0], IndexedArray([3, 2, 4, 2, 2, 4, 0], [0.0, 1.1, 2.2, 3.3, 4.4]))
    #     self.assertEqual([x for x in a], [0.0, 4.4, 2.2, 2.2, 4.4, 2.2, 3.3])
    #     self.assertEqual([a[i] for i in range(len(a))], [0.0, 4.4, 2.2, 2.2, 4.4, 2.2, 3.3])
    #     self.assertEqual(a[:].tolist(), [0.0, 4.4, 2.2, 2.2, 4.4, 2.2, 3.3])

    #     a = IndexedArray([6, 5, 4, 3, 6], IndexedArray([3, 2, 4, 2, 2, 4, 0], [0.0, 1.1, 2.2, 3.3, 4.4]))
    #     self.assertEqual([x for x in a], [0.0, 4.4, 2.2, 2.2, 0.0])
    #     self.assertEqual([a[i] for i in range(len(a))], [0.0, 4.4, 2.2, 2.2, 0.0])
    #     self.assertEqual(a[:].tolist(), [0.0, 4.4, 2.2, 2.2, 0.0])

    # def test_byteindexed_get(self):
    #     a = ByteIndexedArray([12, 8, 4, 0], b"\x00\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00", numpy.int32)
    #     self.assertEqual([x for x in a], [3, 2, 1, 0])
    #     self.assertEqual([a[i] for i in range(len(a))], [3, 2, 1, 0])
    #     self.assertEqual(a[:].tolist(), [3, 2, 1, 0])
    #     self.assertEqual(a[[3, 2, 1, 0]].tolist(), [0, 1, 2, 3])
    #     self.assertEqual(a[[True, False, True, False]].tolist(), [3, 1])

    # def test_byteindexed_get5byte(self):
    #     a = ByteIndexedArray([15, 10, 5, 1], b"\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x02\x00\x00\x00\x00\x03\x00\x00\x00\x00", numpy.int32)
    #     self.assertEqual([x for x in a], [3, 2, 1, 0])
    #     self.assertEqual([a[i] for i in range(len(a))], [3, 2, 1, 0])
    #     self.assertEqual(a[:].tolist(), [3, 2, 1, 0])
    #     self.assertEqual(a[[3, 2, 1, 0]].tolist(), [0, 1, 2, 3])
    #     self.assertEqual(a[[True, False, True, False]].tolist(), [3, 1])

    # def test_indexed_byteindexed(self):
    #     a = IndexedArray([1, 2, 3], ByteIndexedArray([12, 8, 4, 0], b"\x00\x00\x00\x00\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00", numpy.int32))
    #     self.assertEqual([a[i] for i in range(len(a))], [2, 1, 0])
    #     self.assertEqual(a[:].tolist(), [2, 1, 0])

    # def test_union_get(self):
    #     a = UnionArray([0, 1, 0, 1, 0, 1, 0, 1, 0, 1], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [[0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], [0, 100, 200, 300, 400, 500, 600, 700, 800, 900]])
    #     self.assertEqual(a.tolist(), [0.0, 100, 2.2, 300, 4.4, 500, 6.6, 700, 8.8, 900])



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numbers
import types

import awkward.array.base
import awkward.util

class Methods(object):
    @staticmethod
    def mixin(methods, awkwardtype):
        assert issubclass(methods, Methods)
        assert not issubclass(methods, awkward.array.base.AwkwardArray)
        assert issubclass(awkwardtype, awkward.array.base.AwkwardArray)
        assert not issubclass(awkwardtype, Methods)
        return type(awkwardtype.__name__ + "Methods", (methods, awkwardtype), {})

    @staticmethod
    def maybemixin(sample, awkwardtype):
        if issubclass(sample, Methods):
            assert issubclass(sample, awkward.array.base.AwkwardArray)
            assert issubclass(awkwardtype, awkward.array.base.AwkwardArray)
            allbases = tuple(x for x in sample.__bases__ if not issubclass(x, awkward.array.base.AwkwardArray)) + (awkwardtype,)
            return type(awkwardtype.__name__ + "Methods", allbases, {})
        else:
            return awkwardtype

class ObjectArray(awkward.array.base.AwkwardArray):
    def __init__(self, content, generator, *args, **kwargs):
        self.content = content
        self.generator = generator
        self.args = args
        self.kwargs = kwargs

    def copy(self, content=None, generator=None, args=None, kwargs=None):
        out = self.__class__.__new__(self.__class__)
        out._content = self._content
        out._generator = self._generator
        out._args = self._args
        out._kwargs = self._kwargs
        if content is not None:
            out.content = content
        if generator is not None:
            out.generator = generator
        if args is not None:
            out.args = args
        if kwargs is not None:
            out.kwargs = kwargs
        return out

    def deepcopy(self, content=None, generator=None, args=None, kwargs=None):
        out = self.copy(content=content, generator=generator, args=args, kwargs=kwargs)
        out._content = awkward.util.deepcopy(out._content)
        return out

    def empty_like(self, **overrides):
        mine = {}
        mine["generator"] = overrides.pop("generator", self._generator)
        mine["args"] = overrides.pop("args", self._args)
        mine["kwargs"] = overrides.pop("kwargs", self._kwargs)
        if isinstance(self._content, awkward.util.numpy.ndarray):
            return self.copy(content=awkward.util.numpy.empty_like(self._content), **mine)
        else:
            return self.copy(content=self._content.empty_like(**overrides), **mine)

    def zeros_like(self, **overrides):
        mine = {}
        mine["generator"] = overrides.pop("generator", self._generator)
        mine["args"] = overrides.pop("args", self._args)
        mine["kwargs"] = overrides.pop("kwargs", self._kwargs)
        if isinstance(self._content, awkward.util.numpy.ndarray):
            return self.copy(content=awkward.util.numpy.zeros_like(self._content), **mine)
        else:
            return self.copy(content=self._content.zeros_like(**overrides), **mine)

    def ones_like(self, **overrides):
        mine = {}
        mine["generator"] = overrides.pop("generator", self._generator)
        mine["args"] = overrides.pop("args", self._args)
        mine["kwargs"] = overrides.pop("kwargs", self._kwargs)
        if isinstance(self._content, awkward.util.numpy.ndarray):
            return self.copy(content=awkward.util.numpy.ones_like(self._content), **mine)
        else:
            return self.copy(content=self._content.ones_like(**overrides), **mine)

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, value):
        self._content = awkward.util.toarray(value, awkward.util.CHARTYPE, (awkward.util.numpy.ndarray, awkward.array.base.AwkwardArray))

    @property
    def generator(self):
        return self._generator

    @generator.setter
    def generator(self, value):
        if not callable(value):
            raise TypeError("generator must be a callable (of one argument: the array slice)")
        self._generator = value

    @property
    def args(self):
        return self._args

    @args.setter
    def args(self, value):
        if not isinstance(value, tuple):
            value = (value,)
        self._args = value

    @property
    def kwargs(self):
        return self._kwargs

    @kwargs.setter
    def kwargs(self, value):
        if not isinstance(value, dict):
            raise TypeError("kwargs must be a dict")
        self._kwargs = value

    @property
    def type(self):
        out = awkward.type.fromarray(self._content)
        out.to = self._generator
        return out

    def __len__(self):
        return len(self._content)

    @property
    def shape(self):
        return self._content.shape

    @property
    def columns(self):
        return self._content.columns

    @property
    def allcolumns(self):
        return self._content.allcolumns

    @property
    def base(self):
        return self._content.base

    def _argfields(self, function):
        if (isinstance(function, types.FunctionType) and function.__code__.co_argcount == 1) or isinstance(self._content, awkward.util.numpy.ndarray):
            return awkward.util._argfields(function)
        else:
            return self._content._argfields(function)

    @property
    def dtype(self):
        return awkward.util.numpy.dtype(object)

    def __iter__(self):
        for x in self._content:
            yield self.generator(x, *self._args, **self._kwargs)

    def __getitem__(self, where):
        if awkward.util.isstringslice(where):
            return self._content[where]

        if isinstance(where, tuple) and where == ():
            return self
        if not isinstance(where, tuple):
            where = (where,)
        head, tail = where[0], where[1:]

        content = self._content[head]
        if isinstance(head, (numbers.Integral, awkward.util.numpy.integer)):
            if isinstance(tail, tuple) and tail == ():
                return self.generator(content, *self._args, **self._kwargs)
            else:
                return self.generator(content, *self._args, **self._kwargs)[tail]

        elif isinstance(tail, tuple) and tail == ():
            return self.copy(content=content)

        else:
            return [x[tail] for x in content]

    def __setitem__(self, where, what):
        self._content[where] = what

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        if method != "__call__":
            return NotImplemented

        contents = []
        for x in inputs:
            if isinstance(x, ObjectArray):
                contents.append(x._content)
            else:
                contents.append(x)

        result = getattr(ufunc, method)(*contents, **kwargs)

        if awkward.util.iscomparison(ufunc):
            return result
        else:
            return self.copy(content=result)



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import struct
import unittest

import numpy

from awkward import *

class Test(unittest.TestCase):
    def runTest(self):
        pass

    ################### old tests

    # def test_virtual_nocache(self):
    #     a = VirtualArray(lambda: [1, 2, 3])
    #     self.assertFalse(a.ismaterialized)
    #     self.assertTrue(numpy.array_equal(a[:], numpy.array([1, 2, 3])))
    #     self.assertTrue(a.ismaterialized)

    #     a = VirtualArray(lambda: range(10))
    #     self.assertFalse(a.ismaterialized)
    #     self.assertTrue(numpy.array_equal(a[::2], numpy.array([0, 2, 4, 6, 8])))
    #     self.assertTrue(a.ismaterialized)

    #     a = VirtualArray(lambda: range(10))
    #     self.assertFalse(a.ismaterialized)
    #     self.assertTrue(numpy.array_equal(a[[5, 3, 6, 0, 6]], numpy.array([5, 3, 6, 0, 6])))
    #     self.assertTrue(a.ismaterialized)

    #     a = VirtualArray(lambda: range(10))
    #     self.assertFalse(a.ismaterialized)
    #     self.assertTrue(numpy.array_equal(a[[True, False, True, False, True, False, True, False, True, False]], numpy.array([0, 2, 4, 6, 8])))
    #     self.assertTrue(a.ismaterialized)

    # def test_virtual_transientcache(self):
    #     cache = {}
    #     a = VirtualArray(lambda: [1, 2, 3], cache=cache)
    #     self.assertFalse(a.ismaterialized)
    #     a[:]
    #     self.assertTrue(a.ismaterialized)
    #     self.assertEqual(list(cache), [a.TransientKey(id(a))])
    #     self.assertEqual(list(cache), [a.key])
    #     self.assertTrue(numpy.array_equal(cache[a.key], numpy.array([1, 2, 3])))
    #     del a

    # def test_virtual_persistentcache(self):
    #     cache = {}
    #     a = VirtualArray(lambda: [1, 2, 3], cache=cache, persistentkey="find-me-again")
    #     self.assertFalse(a.ismaterialized)
    #     a[:]
    #     self.assertTrue(a.ismaterialized)
    #     self.assertEqual(list(cache), ["find-me-again"])
    #     self.assertEqual(list(cache), [a.key])
    #     self.assertTrue(numpy.array_equal(cache[a.key], numpy.array([1, 2, 3])))
    #     del a

    # def test_virtual_dontmaterialize(self):
    #     a = VirtualArray(lambda: [1, 2, 3], dtype=int, shape=(3,))
    #     self.assertFalse(a.ismaterialized)
    #     self.assertEqual(a.dtype, numpy.dtype(int))
    #     self.assertEqual(a.shape, (3,))
    #     self.assertEqual(len(a), 3)
    #     self.assertEqual(a._array, None)
    #     self.assertFalse(a.ismaterialized)
    #     self.assertTrue(numpy.array_equal(a[:], numpy.array([1, 2, 3])))
    #     self.assertTrue(a.ismaterialized)

    # def test_virtualobject_floats(self):
    #     class Point(object):
    #         def __init__(self, array):
    #             self.x, self.y, self.z = array
    #         def __repr__(self):
    #             return "<Point {0} {1} {2}>".format(self.x, self.y, self.z)
    #         def __eq__(self, other):
    #             return isinstance(other, Point) and self.x == other.x and self.y == other.y and self.z == other.z

    #     a = VirtualObjectArray(Point, [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], [7.7, 8.8, 9.9]])
    #     self.assertEqual(a[0], Point([1.1, 2.2, 3.3]))
    #     self.assertEqual(a[1], Point([4.4, 5.5, 6.6]))
    #     self.assertEqual(a[2], Point([7.7, 8.8, 9.9]))
    #     self.assertEqual(a[:], [Point([1.1, 2.2, 3.3]), Point([4.4, 5.5, 6.6]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[::2], [Point([1.1, 2.2, 3.3]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[[True, False, True]], [Point([1.1, 2.2, 3.3]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[[2, 0]], [Point([7.7, 8.8, 9.9]), Point([1.1, 2.2, 3.3])])

    # def test_virtualobject_bytes(self):
    #     class Point(object):
    #         def __init__(self, bytes):
    #             self.x, self.y, self.z = struct.unpack("ddd", bytes)
    #         def __repr__(self):
    #             return "<Point {0} {1} {2}>".format(self.x, self.y, self.z)
    #         def __eq__(self, other):
    #             return isinstance(other, Point) and self.x == other.x and self.y == other.y and self.z == other.z

    #     a = VirtualObjectArray(Point, numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]).view("u1").reshape(-1, 24))
    #     self.assertEqual(a[0], Point(numpy.array([1.1, 2.2, 3.3]).tobytes()))
    #     self.assertEqual(a[1], Point(numpy.array([4.4, 5.5, 6.6]).tobytes()))
    #     self.assertEqual(a[2], Point(numpy.array([7.7, 8.8, 9.9]).tobytes()))
    #     self.assertEqual(a[:], [Point(numpy.array([1.1, 2.2, 3.3]).tobytes()), Point(numpy.array([4.4, 5.5, 6.6]).tobytes()), Point(numpy.array([7.7, 8.8, 9.9]).tobytes())])
    #     self.assertEqual(a[::2], [Point(numpy.array([1.1, 2.2, 3.3]).tobytes()), Point(numpy.array([7.7, 8.8, 9.9]).tobytes())])
    #     self.assertEqual(a[[True, False, True]], [Point(numpy.array([1.1, 2.2, 3.3]).tobytes()), Point(numpy.array([7.7, 8.8, 9.9]).tobytes())])
    #     self.assertEqual(a[[2, 0]], [Point(numpy.array([7.7, 8.8, 9.9]).tobytes()), Point(numpy.array([1.1, 2.2, 3.3]).tobytes())])

    # def test_virtualobject_indexedbytes(self):
    #     class Point(object):
    #         def __init__(self, array):
    #             self.x, self.y, self.z = array
    #         def __repr__(self):
    #             return "<Point {0} {1} {2}>".format(self.x, self.y, self.z)
    #         def __eq__(self, other):
    #             return isinstance(other, Point) and self.x == other.x and self.y == other.y and self.z == other.z

    #     a = VirtualObjectArray(Point, ByteIndexedArray([0, 24, 48], numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]).view("u1"), numpy.dtype((float, 3))))
    #     self.assertEqual(a[0], Point([1.1, 2.2, 3.3]))
    #     self.assertEqual(a[1], Point([4.4, 5.5, 6.6]))
    #     self.assertEqual(a[2], Point([7.7, 8.8, 9.9]))
    #     self.assertEqual(a[:], [Point([1.1, 2.2, 3.3]), Point([4.4, 5.5, 6.6]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[::2], [Point([1.1, 2.2, 3.3]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[[True, False, True]], [Point([1.1, 2.2, 3.3]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[[2, 0]], [Point([7.7, 8.8, 9.9]), Point([1.1, 2.2, 3.3])])

    # def test_virtualobject_jaggedbytes(self):
    #     class Point(object):
    #         def __init__(self, array):
    #             self.x, self.y, self.z = array
    #         def __repr__(self):
    #             return "<Point {0} {1} {2}>".format(self.x, self.y, self.z)
    #         def __eq__(self, other):
    #             return isinstance(other, Point) and self.x == other.x and self.y == other.y and self.z == other.z

    #     a = VirtualObjectArray(Point, ByteJaggedArray.fromoffsets([0, 24, 48, 72], numpy.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9]).view("u1"), float))
    #     self.assertEqual(a[0], Point([1.1, 2.2, 3.3]))
    #     self.assertEqual(a[1], Point([4.4, 5.5, 6.6]))
    #     self.assertEqual(a[2], Point([7.7, 8.8, 9.9]))
    #     self.assertEqual(a[:], [Point([1.1, 2.2, 3.3]), Point([4.4, 5.5, 6.6]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[::2], [Point([1.1, 2.2, 3.3]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[[True, False, True]], [Point([1.1, 2.2, 3.3]), Point([7.7, 8.8, 9.9])])
    #     self.assertEqual(a[[2, 0]], [Point([7.7, 8.8, 9.9]), Point([1.1, 2.2, 3.3])])



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy

import awkward.util

class AwkwardArray(awkward.util.NDArrayOperatorsMixin):
    def __array__(self, *args, **kwargs):
        raise Exception("{0} {1}".format(args, kwargs))

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def __str__(self):
        if len(self) <= 6:
            return "[{0}]".format(" ".join(str(x) if isinstance(x, (numpy.ndarray, AwkwardArray)) else repr(x) for x in self))
        else:
            return "[{0} ... {1}]".format(
                " ".join(str(x) if isinstance(x, (numpy.ndarray, AwkwardArray)) else repr(x) for x in self[:3]),
                ", ".join(str(x) if isinstance(x, (numpy.ndarray, AwkwardArray)) else repr(x) for x in self[-3:]))

    def __repr__(self):
        return "<{0} {1} at {2:012x}>".format(self.__class__.__name__, str(self), id(self))

    def _try_tolist(self, x):
        try:
            return x.tolist()
        except AttributeError:
            return x

    def __getattr__(self, where):
        if awkward.util.is_intstring(where):
            return self[where[1:]]
        else:
            raise AttributeError("'{0}' object has no attribute '{1}'".format(self.__class__.__name__, where))
    
    def tolist(self):
        import awkward.array.table
        out = []
        for x in self:
            if isinstance(x, awkward.array.table.Table.Row):
                out.append(dict((n, self._try_tolist(x[n])) for n in x._table._content))
            elif isinstance(x, numpy.ma.core.MaskedConstant):
                out.append(None)
            else:
                out.append(self._try_tolist(x))
        return out

    def valid(self):
        try:
            self._valid()
        except:
            return False
        else:
            return True

    def apply(self, function):
        args, kwargs = self._argfields(function)
        if args is None and kwargs is None:
            return function(self)
        else:
            args = tuple(self[n] for n in args)
            kwargs = dict((n, self[n]) for n in kwargs)
            return function(*args, **kwargs)

    def filter(self, function):
        args, kwargs = self._argfields(function)
        if args is None and kwargs is None:
            return self[function(self)]
        else:
            args = tuple(self[n] for n in args)
            kwargs = dict((n, self[n]) for n in kwargs)
            return self[function(*args, **kwargs)]

    def maxby(self, function):
        args, kwargs = self._argfields(function)
        if args is None and kwargs is None:
            return self[function(self).argmax()]
        else:
            args = tuple(self[n] for n in args)
            kwargs = dict((n, self[n]) for n in kwargs)
            return self[function(*args, **kwargs).argmax()]

    def minby(self, function):
        args, kwargs = self._argfields(function)
        if args is None and kwargs is None:
            return self[function(self).argmin()]
        else:
            args = tuple(self[n] for n in args)
            kwargs = dict((n, self[n]) for n in kwargs)
            return self[function(*args, **kwargs).argmin()]



#!/usr/bin/env python

# Copyright (c) 2018, DIANA-HEP
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
# 
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# 
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numbers

import awkward.array.base
import awkward.util

class VirtualArray(awkward.array.base.AwkwardArray):
    def __init__(self, generator, cache=None, persistentkey=None, dtype=None, shape=None):
        raise NotImplementedError

# class VirtualArray(awkward.array.base.AwkwardArray):
#     class TransientKey(object):
#         def __init__(self, id):
#             self._id = id
#         def __repr__(self):
#             return "<VirtualArray.TransientKey {0}>".format(self._id)
#         def __hash__(self):
#             return hash((VirtualArray.TransientKey, self._id))
#         def __eq__(self, other):
#             return isinstance(other, VirtualArray.TransientKey) and self._id == other._id
#         def __ne__(self, other):
#             return not self.__eq__(other)
#         def __getstate__(self):
#             raise RuntimeError("VirtualArray.TransientKeys are not unique across processes, and hence should not be serialized")

#     def __init__(self, generator, cache=None, persistentkey=None, dtype=None, shape=None):
#         self.generator = generator
#         self.cache = cache
#         self.persistentkey = persistentkey
#         self._array = None

#         if dtype is None:
#             self._dtype = dtype
#         else:
#             self._dtype = numpy.dtype(dtype)

#         if shape is None or (isinstance(shape, tuple) and len(shape) != 0 and all(isinstance(x, (numbers.Integral, numpy.integer)) and x >= 0 for x in shape)):
#             self._shape = shape
#         else:
#             raise TypeError("shape must be None (unknown) or a non-empty tuple of non-negative integers")

#     @property
#     def generator(self):
#         return self._generator

#     @generator.setter
#     def generator(self, value):
#         if not callable(value):
#             raise TypeError("generator must be a callable (of zero arguments)")
#         self._generator = value

#     @property
#     def cache(self):
#         return self._cache

#     @cache.setter
#     def cache(self, value):
#         if not value is None and not (callable(getattr(value, "__getitem__", None)) and callable(getattr(value, "__setitem__", None)) and callable(getattr(value, "__delitem__", None))):
#             raise TypeError("cache must be a dict or have __getitem__/__setitem__/__delitem__ methods")
#         self._cache = value

#     @property
#     def persistentkey(self):
#         return self._persistentkey

#     @persistentkey.setter
#     def persistentkey(self, value):
#         if value is not None and not isinstance(value, awkward.util.string):
#             raise TypeError("persistentkey must be a string or None")
#         self._persistentkey = value

#     @property
#     def dtype(self):
#         if self._dtype is not None:
#             return self._dtype
#         else:
#             return self.array.dtype

#     @property
#     def shape(self):
#         if self._shape is not None:
#             return self._shape
#         else:
#             return self.array.shape

#     @property
#     def key(self):
#         if self._persistentkey is not None:
#             return self._persistentkey
#         else:
#             return self.TransientKey(id(self))

#     @property
#     def array(self):
#         # Normal states:   (1) no cache and _array is None: make a new one
#         #                  (2) no cache and _array is an array: return _array
#         #                  (3) have a cache and _array is None: make a new one (filling cache)
#         #                  (4) have a cache and _array is a key and cache[key] was evicted: make a new one (filling cache)
#         #                  (5) have a cache and _array is a key and cache[key] exists: return cache[key]
#         # 
#         # Abnormal states: (6) no cache and _array is a key (user removed _cache): make a new one
#         #                  (7) have a cache and _array is an array (user added _cache): fill cache and return _array

#         something = self._array

#         if something is None:
#             # states (1) and (3)
#             return self.materialize()

#         elif self._cache is None:
#             if isinstance(something, (VirtualArray.TransientKey, awkward.util.string)):
#                 # abnormal state (6)
#                 return self.materialize()
#             else:
#                 # state (2)
#                 return something

#         else:
#             if isinstance(something, (VirtualArray.TransientKey, awkward.util.string)):
#                 try:
#                     # state (5)
#                     return self._cache[something]
#                 except:
#                     # state (4), taking any error in __getitem__ as evidence that it was evicted
#                     return self.materialize()
#             else:
#                 # abnormal state (7)
#                 self._cache[self.key] = something
#                 return something

#     @property
#     def ismaterialized(self):
#         if self._cache is None:
#             return isinstance(self._array, (numpy.ndarray, awkward.array.base.AwkwardArray))
#         else:
#             return self._array is not None and self._array in self._cache

#     def materialize(self):
#         array = self._toarray(self.generator(), self.CHARTYPE, (numpy.ndarray, awkward.array.base.AwkwardArray))

#         if self._dtype is not None and self._dtype != array.dtype:
#             raise ValueError("materialized array has dtype {0}, expected dtype {1}".format(array.dtype, self._dtype))
#         if self._shape is not None and self._shape != array.shape:
#             raise ValueError("materialized array has shape {0}, expected shape {1}".format(array.shape, self._shape))
#         if len(array.shape) == 0:
#             raise ValueError("materialized object is scalar: {0}".format(array))

#         if self._cache is None:
#             # states (1), (2), and (6)
#             self._array = array
#         else:
#             # states (3) and (4)
#             self._array = self.key
#             self._cache[self._array] = array

#         return array

#     def __del__(self):
#         # TransientKeys are based on runtime ids, which Python may reuse after an object is garbage collected
#         # they *MUST* be removed from the cache to avoid confusion; persistentkeys can (and should) stay in
#         if self._cache is not None and isinstance(self._array, VirtualArray.TransientKey):
#             try:
#                 del self._cache[self._array]
#             except:
#                 pass

#     def __len__(self):
#         return self.shape[0]

#     def __getitem__(self, where):
#         return self.array[where]
